#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Fri Jun 15 09:31:10 2018

@author: gtucker
"""

import numpy as np
from numpy.testing import assert_array_equal, assert_equal

from landlab import HexModelGrid
from landlab.ca.boundaries.hex_lattice_tectonicizer import (
    LatticeNormalFault,
    LatticeUplifter,
)
from landlab.ca.celllab_cts import Transition
from landlab.ca.oriented_hex_cts import OrientedHexCTS


def test_links_to_update():
    """Test that update list includes lower 2 rows and fault-crossing links"""

    # Create a 6x6 test grid
    hg = HexModelGrid((6, 6), node_layout="rect", orientation="vertical")

    lnf = LatticeNormalFault(grid=hg, fault_x_intercept=-0.1)

    assert_array_equal(
        lnf.links_to_update,
        [
            6,
            7,
            9,
            10,
            11,
            12,
            13,
            14,
            16,
            17,
            18,
            19,
            20,
            22,
            23,
            24,
            25,
            26,
            27,
            28,
            29,
            30,
            33,
            34,
            35,
            36,
            38,
            39,
            41,
            44,
            49,
            52,
            54,
            58,
            60,
            66,
            68,
            71,
            74,
            75,
            78,
        ],
    )


def test_shift_link_and_transition_data_upward():
    """Test the LatticeUplifter method that uplifts link data and tr'ns."""

    mg = HexModelGrid((4, 3), spacing=1.0, orientation="vertical", node_layout="rect")
    nsd = {0: "yes", 1: "no"}
    xnlist = []
    xnlist.append(Transition((0, 0, 0), (1, 1, 0), 1.0, "frogging"))
    xnlist.append(Transition((0, 0, 1), (1, 1, 1), 1.0, "frogging"))
    xnlist.append(Transition((0, 0, 2), (1, 1, 2), 1.0, "frogging"))
    nsg = mg.add_zeros("node_state_grid", at="node")
    ohcts = OrientedHexCTS(mg, nsd, xnlist, nsg)

    assert_array_equal(
        ohcts.link_state[mg.active_links], [0, 4, 8, 8, 4, 0, 4, 8, 8, 4, 0]
    )

    assert_array_equal(
        ohcts.next_trn_id[mg.active_links], [0, 1, 2, 2, 1, 0, 1, 2, 2, 1, 0]
    )

    assert_array_equal(
        np.round(ohcts.next_update[mg.active_links], 2),
        [0.8, 1.26, 0.92, 0.79, 0.55, 1.04, 0.58, 2.22, 3.31, 0.48, 1.57],
    )

    pq = ohcts.priority_queue

    assert_equal(pq._queue[0][2], 19)  # link for first event = 19, not shifted
    assert_equal(round(pq._queue[0][0], 2), 0.48)  # trn scheduled for t = 0.48
    assert_equal(pq._queue[2][2], 14)  # this event scheduled for link 15...
    assert_equal(round(pq._queue[2][0], 2), 0.58)  # ...trn sched for t = 0.58

    lu = LatticeUplifter(grid=mg)
    lu.shift_link_and_transition_data_upward(ohcts, 0.0)

    # note new events lowest 5 links
    assert_array_equal(
        np.round(ohcts.next_update[mg.active_links], 2),
        [0.75, 0.84, 2.6, 0.07, 0.09, 0.8, 0.02, 1.79, 1.51, 2.04, 3.85],
    )
    assert_equal(pq._queue[0][2], 14)  # new soonest event
    assert_equal(pq._queue[9][2], 13)  # was previously 7, now shifted up...
    assert_equal(round(pq._queue[9][0], 2), 0.8)  # ...still sched for t = 0.80



# -*- coding: utf-8 -*-
"""
Created on Sat Nov 14 10:48:23 2015

@author: gtucker
"""
import numpy as np
from numpy.testing import assert_array_equal

from landlab import RasterModelGrid


def test_link_order():
    """Test ordering of links."""
    rg = RasterModelGrid((3, 4))
    assert_array_equal(
        rg.node_at_link_tail, [0, 1, 2, 0, 1, 2, 3, 4, 5, 6, 4, 5, 6, 7, 8, 9, 10]
    )
    assert_array_equal(
        rg.node_at_link_head, [1, 2, 3, 4, 5, 6, 7, 5, 6, 7, 8, 9, 10, 11, 9, 10, 11]
    )


def test_link_at_face():
    grid = RasterModelGrid((4, 5), xy_spacing=1.0)
    assert_array_equal(
        grid.link_at_face,
        [5, 6, 7, 9, 10, 11, 12, 14, 15, 16, 18, 19, 20, 21, 23, 24, 25],
    )

    grid = RasterModelGrid((3, 4))
    assert_array_equal(grid.link_at_face, [4, 5, 7, 8, 9, 11, 12])


def test_horizontal_links():
    grid = RasterModelGrid((4, 5), xy_spacing=1.0)
    assert_array_equal(
        grid.horizontal_links,
        [0, 1, 2, 3, 9, 10, 11, 12, 18, 19, 20, 21, 27, 28, 29, 30],
    )


def test_vertical_links():
    grid = RasterModelGrid((4, 5), xy_spacing=1.0)
    assert_array_equal(
        grid.vertical_links, [4, 5, 6, 7, 8, 13, 14, 15, 16, 17, 22, 23, 24, 25, 26]
    )


def test_link_dirs_at_node():
    grid = RasterModelGrid((4, 3))
    assert_array_equal(
        grid.link_dirs_at_node,
        [
            [-1, -1, 0, 0],
            [-1, -1, 1, 0],
            [0, -1, 1, 0],
            [-1, -1, 0, 1],
            [-1, -1, 1, 1],
            [0, -1, 1, 1],
            [-1, -1, 0, 1],
            [-1, -1, 1, 1],
            [0, -1, 1, 1],
            [-1, 0, 0, 1],
            [-1, 0, 1, 1],
            [0, 0, 1, 1],
        ],
    )
    assert grid.link_dirs_at_node.dtype == np.int8



#!/usr/env/python

"""fill_sinks_barnes.py.

Fill sinks in a landscape to the brim, following the Barnes et al.
(2014) algorithms.
"""


import numpy as np

from landlab.components import LakeMapperBarnes
from landlab.utils.return_array import return_array_at_node


class SinkFillerBarnes(LakeMapperBarnes):
    """Uses the Barnes et al (2014) algorithms to replace pits in a topography
    with flats, or optionally with very shallow gradient surfaces to allow
    continued draining.

    This component is NOT intended for use iteratively as a model runs;
    rather, it is to fill in an initial topography. If you want to repeatedly
    fill pits as a landscape develops, you are after the LakeMapperBarnes
    component. If you want flow paths on your filled landscape, manually run a
    FlowDirector and FlowAccumulator for yourself.

    The locations and depths etc. of the fills will be tracked, and properties
    are provided to access this information.

    References
    ----------
    **Required Software Citation(s) Specific to this Component**

    Barnes, R., Lehman, C., Mulla, D. (2014). Priority-flood: An optimal
    depression-filling and watershed-labeling algorithm for digital elevation
    models. Computers and Geosciences  62(C), 117 - 127.
    https://dx.doi.org/10.1016/j.cageo.2013.04.024

    **Additional References**

    None Listed

    """

    _name = "SinkFillerBarnes"

    _unit_agnostic = True

    _cite_as = """@article{BARNES2014117,
        title = "Priority-flood: An optimal depression-filling and watershed-labeling algorithm for digital elevation models",
        journal = "Computers & Geosciences",
        volume = "62",
        pages = "117 - 127",
        year = "2014",
        issn = "0098-3004",
        doi = "https://doi.org/10.1016/j.cageo.2013.04.024",
        url = "http://www.sciencedirect.com/science/article/pii/S0098300413001337",
        author = "Richard Barnes and Clarence Lehman and David Mulla",
        keywords = "Pit filling, Terrain analysis, Hydrology, Drainage network, Modeling, GIS"
        }"""

    _info = {
        "sediment_fill__depth": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "Depth of sediment added at eachnode",
        },
        "topographic__elevation": {
            "dtype": float,
            "intent": "inout",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "Land surface topographic elevation",
        },
    }

    def __init__(
        self,
        grid,
        surface="topographic__elevation",
        method="D8",
        fill_flat=False,
        ignore_overfill=False,
    ):
        """Initialise the component.

        Parameters
        ----------
        grid : ModelGrid
            A grid.
        surface : field name at node or array of length node
            The surface to fill.
        method : {'Steepest', 'D8'}
            Whether or not to recognise diagonals as valid flow paths, if a raster.
            Otherwise, no effect.
        fill_flat : bool
            If True, pits will be filled to perfectly horizontal. If False, the new
            surface will be slightly inclined (at machine precision) to give
            steepest descent flow paths to the outlet, once they are calculated.
        ignore_overfill : bool
            If True, suppresses the Error that would normally be raised during
            creation of a gentle incline on a fill surface (i.e., if not
            fill_flat). Typically this would happen on a synthetic DEM where more
            than one outlet is possible at the same elevation. If True, the
            was_there_overfill property can still be used to see if this has
            occurred.

        """
        if "flow__receiver_node" in grid.at_node:
            if grid.at_node["flow__receiver_node"].size != grid.size("node"):
                msg = (
                    "A route-to-multiple flow director has been "
                    "run on this grid. The landlab development team has not "
                    "verified that SinkFillerBarnes is compatible with "
                    "route-to-multiple methods. Please open a GitHub Issue "
                    "to start this process."
                )
                raise NotImplementedError(msg)

        # Most of the functionality of this component is directly inherited
        # from SinkFillerBarnes, so
        super().__init__(
            grid,
            surface=surface,
            method=method,
            fill_flat=fill_flat,
            fill_surface=surface,
            redirect_flow_steepest_descent=False,
            reaccumulate_flow=False,
            ignore_overfill=ignore_overfill,
            track_lakes=True,
        )
        # note we will always track the fills, since we're only doing this
        # once... Likewise, no need for flow routing; this is not going to
        # get used dynamically.
        self._supplied_surface = return_array_at_node(grid, surface).copy()
        # create the only new output field:
        self._sed_fill_depth = self._grid.add_zeros(
            "node", "sediment_fill__depth", clobber=True
        )

    def run_one_step(self):
        """Fills the surface to remove all pits.

        Examples
        --------
        >>> import numpy as np
        >>> from landlab import RasterModelGrid
        >>> from landlab.components import SinkFillerBarnes, FlowAccumulator
        >>> mg = RasterModelGrid((5, 6))
        >>> for edge in ('left', 'top', 'bottom'):
        ...     mg.status_at_node[mg.nodes_at_edge(edge)] = mg.BC_NODE_IS_CLOSED
        >>> z = mg.add_zeros('node', 'topographic__elevation', dtype=float)
        >>> z.reshape(mg.shape)[2, 1:-1] = [2., 1., 0.5, 1.5]
        >>> z.reshape(mg.shape)[1, 1:-1] = [2.1, 1.1, 0.6, 1.6]
        >>> z.reshape(mg.shape)[3, 1:-1] = [2.2, 1.2, 0.7, 1.7]
        >>> z_init = z.copy()
        >>> sfb = SinkFillerBarnes(mg, method='Steepest')  #, surface=z

        TODO: once return_array_at_node is fixed, this example should also
        take surface... GIVE IT surface=z  !!

        >>> sfb.run_one_step()
        >>> z_out = np.array([ 0. ,  0. ,  0. ,  0. ,  0. ,  0. ,
        ...                    0. ,  2.1,  1.5,  1.5,  1.6,  0. ,
        ...                    0. ,  2. ,  1.5,  1.5,  1.5,  0. ,
        ...                    0. ,  2.2,  1.5,  1.5,  1.7,  0. ,
        ...                    0. ,  0. ,  0. ,  0. ,  0. ,  0. ])
        >>> # np.allclose(z, z_out)  ->  True once fixed
        >>> np.all(np.equal(z, z_out))  # those 1.5's are actually a bit > 1.5
        False
        >>> fill_map = np.array([-1, -1, -1, -1, -1, -1,
        ...                      -1, -1, 16, 16, -1, -1,
        ...                      -1, -1, 16, 16, -1, -1,
        ...                      -1, -1, 16, 16, -1, -1,
        ...                      -1, -1, -1, -1, -1, -1])
        >>> np.all(sfb.fill_map == fill_map)
        True
        >>> np.all(sfb.fill_at_node == (sfb.fill_map > -1))  # bool equivalent
        True
        >>> sfb.was_there_overfill  # everything fine with slope adding
        False

        >>> fr = FlowAccumulator(mg, flow_director='D4')  # routing now works
        >>> fr.run_one_step()
        >>> np.all(mg.at_node['flow__sink_flag'][mg.core_nodes] == 0)
        True
        >>> drainage_area = np.array([  0.,   0.,   0.,   0.,   0.,   0.,
        ...                             0.,   1.,   2.,   3.,   1.,   1.,
        ...                             0.,   1.,   4.,   9.,  10.,  10.,
        ...                             0.,   1.,   2.,   1.,   1.,   1.,
        ...                             0.,   0.,   0.,   0.,   0.,   0.])
        >>> np.allclose(mg.at_node['drainage_area'], drainage_area)
        True

        Test two pits and a flat fill:

        >>> from collections import deque
        >>> z[:] = mg.node_x.max() - mg.node_x
        >>> z[[10, 23]] = 1.1  # raise "guard" exit nodes
        >>> z[7] = 2.  # is a lake on its own
        >>> z[9] = 0.5
        >>> z[15] = 0.3
        >>> z[14] = 0.6  # [9, 14, 15] is a lake
        >>> z[22] = 0.9  # a non-contiguous lake node also draining to 16
        >>> z_init = z.copy()
        >>> sfb = SinkFillerBarnes(mg, method='Steepest', fill_flat=True)
        >>> sfb.run_one_step()
        >>> sfb.fill_dict  == {8: deque([7]), 16: deque([15, 9, 14, 22])}
        True
        >>> sfb.number_of_fills
        2
        >>> sfb.fill_outlets == [16, 8]
        True
        >>> np.allclose(sfb.fill_areas, np.array([4., 1.]))  # same order
        True

        Unlike the LakeMapperBarnes equivalents, fill_depths and fill_volumes
        are always available through this component:

        >>> fill_depths = np.array([ 0. ,  0. ,  0. ,  0. ,  0. ,  0. ,
        ...                          0. ,  1. ,  0. ,  0.5,  0. ,  0. ,
        ...                          0. ,  0. ,  0.4,  0.7,  0. ,  0. ,
        ...                          0. ,  0. ,  0. ,  0. ,  0.1,  0. ,
        ...                          0. ,  0. ,  0. ,  0. ,  0. ,  0. ])
        >>> np.allclose(sfb.fill_depths, fill_depths)
        True
        >>> np.allclose(sfb.fill_volumes, np.array([1.7, 1.]))
        True

        Note that with a flat fill, we can't drain the surface. The surface
        is completely flat, so each and every core node within the fill
        becomes a sink.

        >>> fr.run_one_step()
        >>> where_is_filled = np.where(sfb.fill_map > -1, 1, 0)
        >>> np.all(
        ...     mg.at_node['flow__sink_flag'][mg.core_nodes] ==
        ...     where_is_filled[mg.core_nodes])
        True

        (Note that the fill_map does not think that the perimeter nodes are
        sinks, since they haven't changed elevation. In contrast, the
        FlowAccumulator *does* think they are, because these nodes are where
        flow terminates.)
        """
        if "flow__receiver_node" in self._grid.at_node:
            if self._grid.at_node["flow__receiver_node"].size != self._grid.size(
                "node"
            ):
                msg = (
                    "A route-to-multiple flow director has been "
                    "run on this grid. The landlab development team has "
                    "not verified that SinkFillerBarnes is compatible with "
                    "route-to-multiple methods. Please open a GitHub Issue "
                    "to start this process."
                )
                raise NotImplementedError(msg)

        super().run_one_step()

        self._sed_fill_depth[:] = self._surface - self._supplied_surface

    @property
    def fill_dict(self):
        """Return a dictionary where the keys are the outlet nodes of each
        filled area, and the values are deques of nodes within each.

        Items are not returned in ID order.
        """
        return super().lake_dict

    @property
    def fill_outlets(self):
        """Returns the outlet for each filled area, not necessarily in ID
        order."""
        return super().lake_outlets

    @property
    def number_of_fills(self):
        """Return the number of individual filled areas."""
        return super().number_of_lakes

    @property
    def fill_map(self):
        """Return an array of ints, where each filled node is labelled with its
        outlet node ID.

        Nodes not in a filled area are labelled with
        BAD_INDEX_VALUE (default -1).
        """
        return super().lake_map

    @property
    def fill_at_node(self):
        """Return a boolean array, True if the node is filled, False
        otherwise."""
        return super().lake_at_node

    @property
    def fill_depths(self):
        """Return the change in surface elevation at each node this step."""
        return self._sed_fill_depth

    @property
    def fill_areas(self):
        """A nlakes-long array of the area of each fill.

        The order is the same as that of the keys in fill_dict, and of
        fill_outlets.
        """
        return super().lake_areas

    @property
    def fill_volumes(self):
        """A nlakes-long array of the volume of each fill.

        The order is the same as that of the keys in fill_dict, and of
        fill_outlets.
        """
        fill_vols = np.empty(self.number_of_fills, dtype=float)
        col_vols = self._grid.cell_area_at_node * self._sed_fill_depth
        for (i, (outlet, fillnodes)) in enumerate(self.fill_dict.items()):
            fill_vols[i] = col_vols[fillnodes].sum()
        return fill_vols



import numpy as np
from pytest import approx

from landlab import RasterModelGrid

field_dtype = float


def test_zeros(graph_element):
    grid = RasterModelGrid((4, 5))
    number_of_elements = grid.number_of_elements(graph_element)
    assert np.all(
        grid.zeros(at=graph_element, dtype=field_dtype)
        == approx(np.zeros(number_of_elements, dtype=field_dtype))
    )


def test_add_zeros(graph_element):
    grid = RasterModelGrid((4, 5))
    number_of_elements = grid.number_of_elements(graph_element)
    rtn_values = grid.add_zeros("name", at=graph_element, dtype=field_dtype)
    assert rtn_values is grid.field_values(graph_element, "name")
    assert np.all(rtn_values == approx(np.zeros(number_of_elements, dtype=field_dtype)))


def test_ones(graph_element):
    grid = RasterModelGrid((4, 5))
    number_of_elements = grid.number_of_elements(graph_element)
    assert np.all(
        grid.ones(at=graph_element)
        == approx(np.ones(number_of_elements, dtype=field_dtype))
    )


def test_add_ones(graph_element):
    grid = RasterModelGrid((4, 5))
    number_of_elements = grid.number_of_elements(graph_element)
    rtn_values = grid.add_ones("name", at=graph_element, dtype=field_dtype)
    assert rtn_values is grid.field_values(graph_element, "name")
    assert np.all(rtn_values == approx(np.ones(number_of_elements, dtype=field_dtype)))


def test_empty(graph_element):
    grid = RasterModelGrid((4, 5))
    number_of_elements = grid.number_of_elements(graph_element)
    assert grid.empty(at=graph_element, dtype=field_dtype).size == number_of_elements


def test_add_empty(graph_element):
    grid = RasterModelGrid((4, 5))
    rtn_values = grid.add_empty("name", at=graph_element, dtype=field_dtype)
    assert rtn_values is grid.field_values(graph_element, "name")



import inspect

import numpy as np
import pytest
from pytest import approx
from scipy.spatial import Delaunay, Voronoi

from landlab.graph.voronoi.voronoi_to_graph import (
    VoronoiDelaunay,
    VoronoiDelaunayToGraph,
)

XY_OF_NODE = {
    "rect-horizontal-3-3": [
        [0.0, 0.0],
        [1.0, 0.0],
        [2.0, 0.0],
        [0.5, 1.0],
        [1.5, 1.0],
        [2.5, 1.0],
        [0.0, 2.0],
        [1.0, 2.0],
        [2.0, 2.0],
    ],
    "rect-horizontal-4-3": [
        [0.0, 0.0],
        [1.0, 0.0],
        [2.0, 0.0],
        [0.5, 1.0],
        [1.5, 1.0],
        [2.5, 1.0],
        [0.0, 2.0],
        [1.0, 2.0],
        [2.0, 2.0],
        [0.0, 3.0],
        [1.0, 3.0],
        [2.0, 3.0],
    ],
    "rect-vertical-3-3": [
        [0.0, 0.0],
        [2.0, 0.0],
        [1.0, 0.5],
        [0.0, 1.0],
        [2.0, 1.0],
        [1.0, 1.5],
        [0.0, 2.0],
        [2.0, 2.0],
        [1.0, 2.5],
    ],
    "rect-vertical-3-4": [
        [0.0, 0.0],
        [2.0, 0.0],
        [1.0, 0.5],
        [3.0, 0.5],
        [0.0, 1.0],
        [2.0, 1.0],
        [1.0, 1.5],
        [3.0, 1.5],
        [0.0, 2.0],
        [2.0, 2.0],
        [1.0, 2.5],
        [3.0, 2.5],
    ],
}


@pytest.fixture
def hex_graph():
    return VoronoiDelaunayToGraph(XY_OF_NODE["rect-horizontal-3-3"])


@pytest.fixture
def xy_of_hex():
    return XY_OF_NODE["rect-horizontal-3-3"]


def pytest_generate_tests(metafunc):
    if "at_property" in metafunc.fixturenames:
        props = dict(
            inspect.getmembers(
                VoronoiDelaunayToGraph, lambda o: isinstance(o, property)
            )
        )
        metafunc.parametrize(
            "at_property", [name for name in props.keys() if "_at_" in name]
        )
    if "of_property" in metafunc.fixturenames:
        props = dict(
            inspect.getmembers(
                VoronoiDelaunayToGraph, lambda o: isinstance(o, property)
            )
        )
        metafunc.parametrize(
            "of_property",
            [
                name
                for name in props.keys()
                if "_of_" in name and not name.startswith("number")
            ],
        )


def test_voronoi_name_mapping(xy_of_hex):
    """Test scipy Voronoi names are mapped to landlab-style names."""
    voronoi = Voronoi(xy_of_hex)
    delaunay = Delaunay(xy_of_hex)
    graph = VoronoiDelaunay(xy_of_hex)

    assert np.all(graph.x_of_node == approx(voronoi.points[:, 0]))
    assert np.all(graph.y_of_node == approx(voronoi.points[:, 1]))

    assert np.all(graph.x_of_corner == approx(voronoi.vertices[:, 0]))
    assert np.all(graph.y_of_corner == approx(voronoi.vertices[:, 1]))

    assert np.all(graph.nodes_at_link == voronoi.ridge_points)

    assert tuple(graph.n_corners_at_cell) == tuple(
        len(region) for region in voronoi.regions
    )
    for cell, corners in enumerate(graph.corners_at_cell):
        assert np.all(corners[: graph.n_corners_at_cell[cell]] == voronoi.regions[cell])
        assert np.all(corners[graph.n_corners_at_cell[cell] :] == -1)
    assert np.all(graph.corners_at_face == voronoi.ridge_vertices)
    assert np.all(graph.nodes_at_face == voronoi.ridge_points)
    assert np.all(graph.cell_at_node == voronoi.point_region)

    assert np.all(graph.nodes_at_patch == delaunay.simplices)


def test_at_array_is_int(hex_graph, at_property):
    """Test that _at_ properties are arrays of int."""
    assert getattr(hex_graph, at_property).dtype == int


def test_degenerate_case():
    xy_of_node = np.array(
        [[0, 0], [1, 0], [2, 0], [0, 1], [1, 1], [2, 1], [0, 2], [1, 2], [2, 2]],
        dtype=float,
    )
    VoronoiDelaunay(xy_of_node)
    # VoronoiDelaunayToGraph(xy_of_node)


def test_of_array_is_float(hex_graph, of_property):
    """Test that _of_ properties are arrays of float."""
    xy_of_node = np.array(
        [[0, 0], [2, 0], [4, 0], [1, 1], [3, 1], [5, 1], [0, 2], [2, 2], [4, 2]],
        dtype=float,
    )
    # hex_graph = VoronoiDelaunayToGraph(xy_of_node)
    hex_graph = VoronoiDelaunay(xy_of_node)
    assert getattr(hex_graph, of_property).dtype == float


@pytest.mark.parametrize(
    "element,expected",
    [
        ("nodes", 9),
        ("links", 17),
        ("patches", 9),
        ("corners", 9),
        ("faces", 10),
        ("cells", 2),
    ],
)
def test_element_count_without_perimeter_nodes(hex_graph, element, expected):
    assert getattr(hex_graph, "number_of_{0}".format(element)) == expected


@pytest.mark.parametrize(
    "element,expected",
    [
        ("nodes", 9),
        ("links", 16),
        ("patches", 8),
        ("corners", 8),
        ("faces", 8),
        ("cells", 1),
    ],
)
def test_element_count_with_perimeter_nodes(xy_of_hex, element, expected):
    perimeter_links = [[0, 1], [1, 2], [2, 5], [5, 8], [8, 7], [7, 6], [6, 3], [3, 0]]
    graph = VoronoiDelaunayToGraph(xy_of_hex, perimeter_links=perimeter_links)
    assert getattr(graph, "number_of_{0}".format(element)) == expected


@pytest.mark.parametrize("at", ("node", "link", "cell", "corner", "face", "cell"))
def test_compact_ids_without_perimeter_nodes(hex_graph, at):
    ids = []
    for name in hex_graph.ids_with_prefix(at):
        ids.append(np.sort(np.unique(getattr(hex_graph, name).reshape((-1,)))))
    ids = np.sort(np.unique(np.concatenate(ids)))
    ids = ids[ids >= 0]

    assert ids[0] >= 0
    assert ids[-1] <= hex_graph._mesh.dims[at]


@pytest.mark.parametrize("at", ("node", "link", "cell", "corner", "face", "cell"))
def test_compact_ids_with_perimeter_nodes(xy_of_hex, at):
    perimeter_links = [[0, 1], [1, 2], [2, 5], [5, 8], [8, 7], [7, 6], [6, 3], [3, 0]]
    graph = VoronoiDelaunayToGraph(xy_of_hex, perimeter_links=perimeter_links)

    ids = []
    for name in graph.ids_with_prefix(at):
        ids.append(np.sort(np.unique(getattr(graph, name).reshape((-1,)))))
    ids = np.sort(np.unique(np.concatenate(ids)))
    ids = ids[ids >= 0]

    assert ids[0] >= 0
    assert ids[-1] <= graph._mesh.dims[at]


@pytest.mark.parametrize("at", ["node", "link", "patch", "corner", "face", "cell"])
def test_has_prefix(hex_graph, at):
    expected = {
        "node": ("nodes_at_patch", "nodes_at_face", "node_at_cell", "nodes_at_link"),
        "link": ("links_at_patch",),
        "patch": (),
        "corner": ("corners_at_face", "corners_at_cell"),
        "face": ("faces_at_cell",),
        "cell": ("cell_at_node",),
    }

    assert hex_graph.ids_with_prefix(at) == set(expected[at])


@pytest.mark.parametrize("at", ["node", "link", "patch", "corner", "face", "cell"])
def test_has_suffix(hex_graph, at):
    expected = {
        "node": ("cell_at_node",),
        "link": ("nodes_at_link",),
        "patch": ("nodes_at_patch", "links_at_patch"),
        "corner": (),
        "face": ("corners_at_face", "nodes_at_face"),
        "cell": (
            "n_corners_at_cell",
            "faces_at_cell",
            "node_at_cell",
            "corners_at_cell",
        ),
    }

    assert hex_graph.ids_with_suffix(at) == set(expected[at])


@pytest.mark.parametrize(
    "n_nodes",
    [2 ** 10, 2 ** 11, 2 ** 12, 2 ** 13, 2 ** 14, 2 ** 15],  # , 2 ** 16, 2 ** 20]
)
def test_big_graph(n_nodes):
    xy_of_node = np.random.rand(2 * n_nodes).reshape((-1, 2))
    graph = VoronoiDelaunayToGraph(xy_of_node)
    assert graph.number_of_nodes == n_nodes



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jun  4 16:26:31 2019

@author: G Tucker, D Litwin
"""

import numpy as np
from numpy.testing import assert_almost_equal, assert_equal

from landlab import HexModelGrid, RasterModelGrid
from landlab.components import FlowAccumulator, GroundwaterDupuitPercolator
from landlab.grid.mappers import map_mean_of_link_nodes_to_link


def test_simple_water_table():
    """Test a one-node steady simulation.

    Notes
    -----
    The analytical solution for one interior cell with one open boundary is as
    follows. The incoming recharge must equal the outgoing discharge. The
    incoming recharge is R, and the surface area is 1 m2, so the incoming
    volume per unit time is R m/s x 1 m x 1 m. The outgoing discharge is equal
    to the conductivity, K (m/s), times the thickness at the boundary, Hb,
    times the hydraulic gradient, which in this case is (H - 0) / dx = H.
    The model uses the upwind thickness, which in this case is H. Therefore:

        K H^2 = R, or

        H = sqrt( R / K ).

    With R = 10^-8 m/s and K = 10^-2 m/s, we should have

        H = 0.001 m.
    """
    boundaries = {"top": "closed", "left": "closed", "bottom": "closed"}
    rg = RasterModelGrid((3, 3), bc=boundaries)
    rg.add_zeros("aquifer_base__elevation", at="node")
    rg.add_ones("topographic__elevation", at="node")
    gdp = GroundwaterDupuitPercolator(
        rg, recharge_rate=1.0e-8, hydraulic_conductivity=0.01
    )
    for i in range(100):
        gdp.run_one_step(1e3)

    assert_equal(np.round(gdp._thickness[4], 5), 0.001)


def test_simple_surface_leakage():
    """test a one-node steady simulation for surface leakage.

    Notes
    ----
    This test demonstrates that at steady state when no flow is
    allowed to leave the domain through the subsurface, the surface
    water flux is equal to the recharge flux.

    """
    grid = RasterModelGrid((3, 3), xy_spacing=1.0)
    grid.set_closed_boundaries_at_grid_edges(True, True, True, True)
    grid.add_zeros("aquifer_base__elevation", at="node")
    grid.add_ones("topographic__elevation", at="node")
    gdp = GroundwaterDupuitPercolator(grid, recharge_rate=1.0e-6)

    for i in range(1000):
        gdp.run_one_step(1e3)

    assert_almost_equal(gdp._qs[4], 1e-6)


def test_simple_water_table_adaptive_dt():
    """Test a one-node steady simulation.

    Notes
    -----
    This test demonstrates the same simple water table as
    test_simple_water_table, but with the run_with_adaptive_time_step_solver
    method.


    """
    boundaries = {"top": "closed", "left": "closed", "bottom": "closed"}
    rg = RasterModelGrid((3, 3), bc=boundaries)
    rg.add_zeros("aquifer_base__elevation", at="node")
    rg.add_ones("topographic__elevation", at="node")
    rg.add_zeros("water_table__elevation", at="node")
    rg.at_node["water_table__elevation"][rg.core_nodes] += 1e-10
    gdp = GroundwaterDupuitPercolator(
        rg,
        recharge_rate=1.0e-8,
        hydraulic_conductivity=0.01,
    )
    for i in range(10):
        gdp.run_with_adaptive_time_step_solver(1e4)

    assert_equal(np.round(gdp._thickness[4], 5), 0.001)


def test_conservation_of_mass_adaptive_dt():
    """test conservation of mass in a sloping aquifer.

    Notes
    ----
    This test demonstrates conservation of mass from a sloping aquifer
    with constant recharge using the flux and storage calculation methods.
    Note that there is a slight loss of mass as the adaptive timestep solver
    does not calculate fluxes in the intermediate substeps, so accuracy in
    this case is only to ~3 significant digits.
    """

    grid = RasterModelGrid((3, 10), xy_spacing=10.0)
    grid.set_closed_boundaries_at_grid_edges(True, True, False, True)
    elev = grid.add_zeros("topographic__elevation", at="node")
    grid.add_zeros("aquifer_base__elevation", at="node")

    elev[:] = grid.x_of_node / 100 + 1
    wt = grid.add_zeros("water_table__elevation", at="node")
    wt[:] = elev

    # initialize the groundwater model
    gdp = GroundwaterDupuitPercolator(
        grid,
        hydraulic_conductivity=0.0005,
        recharge_rate=1e-7,
        courant_coefficient=0.01,
    )
    fa = FlowAccumulator(grid, runoff_rate="surface_water__specific_discharge")

    # initialize fluxes we will record
    recharge_flux = 0
    gw_flux = 0
    sw_flux = 0
    storage_0 = gdp.calc_total_storage()

    dt = 1e4
    for i in range(500):
        gdp.run_with_adaptive_time_step_solver(dt)
        fa.run_one_step()

        recharge_flux += gdp.calc_recharge_flux_in() * dt
        gw_flux += gdp.calc_gw_flux_out() * dt
        sw_flux += gdp.calc_sw_flux_out() * dt
    storage = gdp.calc_total_storage()

    assert_almost_equal(
        (gw_flux + sw_flux + storage - storage_0) / recharge_flux, 1.0, decimal=3
    )


def test_symmetry_of_solution():
    """test that water table is symmetric under constant recharge

    Notes:
    ----
    Under constant recharge with radially symmetric aquifer base elevation,
    the model should produce a water table that is radially symmetric. This test
    demonstrates this is the case where topographic elevation and aquifer
    base elevation are radially symmetric parabolas on a hexagonal grid.

    """
    hmg = HexModelGrid(shape=(7, 4), spacing=10.0)
    x = hmg.x_of_node
    y = hmg.y_of_node
    elev = hmg.add_zeros("topographic__elevation", at="node")
    elev[:] = 1e-3 * (x * (max(x) - x) + y * (max(y) - y)) + 2
    base = hmg.add_zeros("aquifer_base__elevation", at="node")
    base[:] = elev - 2
    wt = hmg.add_zeros("water_table__elevation", at="node")
    wt[:] = elev
    wt[hmg.open_boundary_nodes] = 0.0

    gdp = GroundwaterDupuitPercolator(
        hmg, recharge_rate=1e-7, hydraulic_conductivity=1e-4
    )
    for i in range(1000):
        gdp.run_one_step(1e3)

    tc = hmg.at_node["aquifer__thickness"]
    assert_almost_equal(tc[5], tc[31])  # SW-NE
    assert_almost_equal(tc[29], tc[7])  # NW-SE
    assert_almost_equal(tc[16], tc[20])  # W-E


def test_wt_above_surface_standard_run_step():
    """test that water tables above the topogrpahic elevation are
    set to the topographic elevation.

    Notes:
    ----
    Water tables above the land surface represent a non-physical condition.
    The GroundwaterDupuitPercolator will not produce this state when it
    is the only component operating on water table elevation or topogrpahic
    elevation, however, when combined with components that do, this may occur.
    If the water table is above the ground surface at a node, it is set to the ground
    surface elevation at that node.

    """

    grid = RasterModelGrid((3, 3))
    grid.set_closed_boundaries_at_grid_edges(True, True, True, False)
    wt = grid.add_ones("node", "water_table__elevation")
    _ = grid.add_ones("node", "topographic__elevation")
    _ = grid.add_zeros("node", "aquifer_base__elevation")

    # initialize the groundwater model
    gdp = GroundwaterDupuitPercolator(grid, recharge_rate=0.0)
    gdp.run_one_step(1)
    assert_equal(wt[4], 1)


def test_wt_above_surface_adaptive_run_step():
    grid = RasterModelGrid((3, 3))
    grid.set_closed_boundaries_at_grid_edges(True, True, True, False)
    wt = grid.add_ones("node", "water_table__elevation")
    _ = grid.add_ones("node", "topographic__elevation")
    _ = grid.add_zeros("node", "aquifer_base__elevation")

    # initialize the groundwater model
    gdp = GroundwaterDupuitPercolator(grid, recharge_rate=0.0)
    gdp.run_with_adaptive_time_step_solver(1)
    assert_equal(wt[4], 1)


def test_inactive_interior_node():

    """
    Test that component returns correct values for recharge flux and
    storage when an interior node is INACTIVE

    Notes:
    ----
    When an interior node is inactive, the number of core nodes is not
    equal to the number of cells. This test confirms that the methods
    to calculate recharge flux and active storage acknowledge this difference.

    """

    mg = RasterModelGrid((4, 4), xy_spacing=1.0)
    mg.status_at_node[5] = mg.BC_NODE_IS_FIXED_VALUE
    elev = mg.add_zeros("node", "topographic__elevation")
    elev[:] = 1
    base = mg.add_zeros("node", "aquifer_base__elevation")
    base[:] = 0
    wt = mg.add_zeros("node", "water_table__elevation")
    wt[:] = 1

    gdp = GroundwaterDupuitPercolator(mg)
    assert_almost_equal(gdp.calc_recharge_flux_in(), 3e-8)
    assert_almost_equal(gdp.calc_total_storage(), 0.6)


def test_k_func():
    """
    Test the use of a function to specify how hydraulic conductivity changes
    with water table position.

    Note:
    ----
    Test that component keeps hydraulic conductivity at default value when
    k_func is None. Then test that a simple function correctly sets the
    hydraulic conductivity value after run_one_step and run_with_adaptive_time_step_solver.

    """

    # initialize model grid
    mg = RasterModelGrid((4, 4), xy_spacing=1.0)
    elev = mg.add_zeros("node", "topographic__elevation")
    elev[:] = 1
    base = mg.add_zeros("node", "aquifer_base__elevation")
    base[:] = 0
    wt = mg.add_zeros("node", "water_table__elevation")
    wt[:] = 0.5

    # initialize model without giving k_func
    gdp = GroundwaterDupuitPercolator(mg)

    # run model and assert that K hasn't changed from the default value
    gdp.run_one_step(0)
    assert np.equal(0.001, gdp.K).all()

    gdp.run_with_adaptive_time_step_solver(0)
    assert np.equal(0.001, gdp.K).all()

    # create a simple k_func, where hydraulic conductivity varies linearly
    # with depth, from Ks at surface to 0 at aquifer base
    def k_func_test(grid, Ks=0.01):
        h = grid.at_node["aquifer__thickness"]
        b = (
            grid.at_node["topographic__elevation"]
            - grid.at_node["aquifer_base__elevation"]
        )
        blink = map_mean_of_link_nodes_to_link(grid, b)
        hlink = map_mean_of_link_nodes_to_link(grid, h)

        return (hlink / blink) * Ks

    # initialize model with given k_func
    gdp1 = GroundwaterDupuitPercolator(mg, hydraulic_conductivity=k_func_test)

    # run model and assert that K has been updated correctly
    gdp1.run_one_step(0)
    assert np.equal(0.005, gdp1.K).all()

    gdp1.run_with_adaptive_time_step_solver(0)
    assert np.equal(0.005, gdp1.K).all()


def test_callback_func():
    """
    Test the use of a callback function to return the storage and
    substep durations while using the run_with_adaptive_time_step_solver
    method.

    Notes:
    ----
    Two tests here: make sure that the substeps sum to the global timestep,
    and make sure that when recharge is 0.0, the total storage does not
    increase during any of the substeps. See component documentation for
    more detail on arguments for the callback_fun.
    """

    # make a function that writes storage and substep duration to
    # externally defined lists
    storage_subdt = []
    subdt = []
    all_n = []

    def test_fun(grid, recharge, dt, n=0.2):
        cores = grid.core_nodes
        h = grid.at_node["aquifer__thickness"]
        area = grid.cell_area_at_node
        storage = np.sum(n * h[cores] * area[cores])

        storage_subdt.append(storage)
        subdt.append(dt)
        all_n.append(n)

    # initialize grid
    grid = RasterModelGrid((3, 3))
    grid.set_closed_boundaries_at_grid_edges(True, True, False, True)
    elev = grid.add_ones("topographic__elevation", at="node")
    elev[3] = 0.1
    grid.add_zeros("aquifer_base__elevation", at="node")
    wt = grid.add_zeros("water_table__elevation", at="node")
    wt[:] = elev

    # initialize groundwater model
    gdp = GroundwaterDupuitPercolator(
        grid,
        recharge_rate=0.0,
        hydraulic_conductivity=0.0001,
        callback_fun=test_fun,
        n=0.1,
    )

    # run groundawter model
    gdp.run_with_adaptive_time_step_solver(1e5)

    # assert that the water table does not increase during substeps
    assert (np.diff(storage_subdt) <= 0.0).all()

    # assert that substeps sum to the global timestep
    assert_almost_equal(1e5, sum(subdt))

    assert all(x == 0.1 for x in all_n)



from .bedrock_landslider import BedrockLandslider
from .chi_index import ChiFinder
from .depression_finder import DepressionFinderAndRouter
from .depth_dependent_diffusion import DepthDependentDiffuser
from .depth_dependent_taylor_soil_creep import DepthDependentTaylorDiffuser
from .detachment_ltd_erosion import DepthSlopeProductErosion, DetachmentLtdErosion
from .diffusion import LinearDiffuser
from .discharge_diffuser import DischargeDiffuser
from .drainage_density import DrainageDensity
from .erosion_deposition import ErosionDeposition
from .fire_generator import FireGenerator
from .flexure import Flexure, Flexure1D
from .flow_accum import FlowAccumulator, LossyFlowAccumulator
from .flow_director import (
    FlowDirectorD8,
    FlowDirectorDINF,
    FlowDirectorMFD,
    FlowDirectorSteepest,
)
from .fracture_grid import FractureGridGenerator
from .gflex import gFlex
from .groundwater import GroundwaterDupuitPercolator
from .hack_calculator import HackCalculator
from .hand_calculator import HeightAboveDrainageCalculator
from .lake_fill import LakeMapperBarnes
from .landslides import LandslideProbability
from .lateral_erosion import LateralEroder
from .lithology import LithoLayers, Lithology
from .marine_sediment_transport import SimpleSubmarineDiffuser
from .network_sediment_transporter import NetworkSedimentTransporter
from .nonlinear_diffusion import PerronNLDiffuse
from .normal_fault import NormalFault
from .overland_flow import (
    KinematicWaveRengers,
    KinwaveImplicitOverlandFlow,
    KinwaveOverlandFlowModel,
    OverlandFlow,
    OverlandFlowBates,
)
from .pet import PotentialEvapotranspiration
from .plant_competition_ca import VegCA
from .potentiality_flowrouting import PotentialityFlowRouter
from .priority_flood_flow_router import PriorityFloodFlowRouter
from .profiler import ChannelProfiler, Profiler, TrickleDownProfiler
from .radiation import Radiation
from .sink_fill import SinkFiller, SinkFillerBarnes
from .soil_moisture import SoilInfiltrationGreenAmpt, SoilMoisture
from .space import Space, SpaceLargeScaleEroder
from .spatial_precip import SpatialPrecipitationDistribution
from .species_evolution import SpeciesEvolver
from .steepness_index import SteepnessFinder
from .stream_power import (
    FastscapeEroder,
    SedDepEroder,
    StreamPowerEroder,
    StreamPowerSmoothThresholdEroder,
)
from .taylor_nonlinear_hillslope_flux import TaylorNonLinearDiffuser
from .tectonics import ListricKinematicExtender
from .tidal_flow import TidalFlowCalculator
from .transport_length_diffusion import TransportLengthHillslopeDiffuser
from .uniform_precip import PrecipitationDistribution
from .vegetation_dynamics import Vegetation
from .weathering import ExponentialWeatherer, ExponentialWeathererIntegrated

COMPONENTS = [
    BedrockLandslider,
    ChannelProfiler,
    ChiFinder,
    DepressionFinderAndRouter,
    DepthDependentDiffuser,
    DepthDependentTaylorDiffuser,
    DepthSlopeProductErosion,
    DetachmentLtdErosion,
    DischargeDiffuser,
    DrainageDensity,
    ErosionDeposition,
    ExponentialWeatherer,
    ExponentialWeathererIntegrated,
    FastscapeEroder,
    FireGenerator,
    Flexure,
    Flexure1D,
    FlowAccumulator,
    PriorityFloodFlowRouter,
    FlowDirectorD8,
    FlowDirectorDINF,
    FlowDirectorMFD,
    FlowDirectorSteepest,
    FractureGridGenerator,
    gFlex,
    GroundwaterDupuitPercolator,
    HackCalculator,
    HeightAboveDrainageCalculator,
    KinematicWaveRengers,
    KinwaveImplicitOverlandFlow,
    KinwaveOverlandFlowModel,
    LakeMapperBarnes,
    LandslideProbability,
    LateralEroder,
    LinearDiffuser,
    ListricKinematicExtender,
    LithoLayers,
    Lithology,
    LossyFlowAccumulator,
    NetworkSedimentTransporter,
    NormalFault,
    OverlandFlow,
    OverlandFlowBates,
    PerronNLDiffuse,
    PotentialEvapotranspiration,
    PotentialityFlowRouter,
    PrecipitationDistribution,
    Profiler,
    Radiation,
    SedDepEroder,
    SimpleSubmarineDiffuser,
    SinkFiller,
    SinkFillerBarnes,
    SoilMoisture,
    SoilInfiltrationGreenAmpt,
    Space,
    SpaceLargeScaleEroder,
    SpatialPrecipitationDistribution,
    SpeciesEvolver,
    SteepnessFinder,
    StreamPowerEroder,
    StreamPowerSmoothThresholdEroder,
    TaylorNonLinearDiffuser,
    TidalFlowCalculator,
    TransportLengthHillslopeDiffuser,
    TrickleDownProfiler,
    VegCA,
    Vegetation,
]

__all__ = [cls.__name__ for cls in COMPONENTS]



#! /usr/env/python
"""Fastscape stream power erosion."""

# This module attempts to "component-ify" GT's Fastscape stream
# power erosion.
# Created DEJH, March 2014.


import numpy as np

from landlab import Component, RasterModelGrid
from landlab.utils.return_array import return_array_at_node

from ..depression_finder.lake_mapper import _FLOODED
from .cfuncs import (
    brent_method_erode_fixed_threshold,
    brent_method_erode_variable_threshold,
)


class FastscapeEroder(Component):

    r"""Fastscape stream power erosion.

    This class uses the Braun-Willett Fastscape approach to calculate the
    amount of erosion at each node in a grid, following a stream power
    framework. This should allow it to be stable against larger timesteps
    than an explicit stream power scheme.

    Note that although this scheme is nominally implicit, and will reach a
    numerically-correct solution under topographic steady state regardless of
    timestep length, the accuracy of transient solutions is *not* timestep
    independent (see Braun & Willett 2013, Appendix B for further details).
    Although the scheme remains significantly more robust and permits longer
    timesteps than a traditional explicit solver under such conditions, it
    is still possible to create numerical instability through use of too long
    a timestep while using this component. The user is cautioned to check their
    implementation is behaving stably before fully trusting it.

    Stream power erosion is implemented as:

    .. math::

        E = K  A ^ m  S ^ n -
               \textit{threshold_sp}

    if :math:`K A ^ m S ^ n > \textit{threshold_sp}`, and:

    .. math:: E = 0,

    if :math:`K A^m S^n <= \textit{threshold_sp}`.

    This module assumes you have already run
    :func:`landlab.components.flow_accum.flow_accumulator.FlowAccumulator.run_one_step`
    in the same timestep. It looks for 'flow__upstream_node_order',
    'flow__link_to_receiver_node', 'drainage_area', 'flow__receiver_node', and
    'topographic__elevation' at the nodes in the grid. 'drainage_area' should
    be in area upstream, not volume (i.e., set runoff_rate=1.0 when calling
    FlowAccumulator.run_one_step).

    The primary method of this class is :func:`run_one_step`.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> from landlab.components import FlowAccumulator, FastscapeEroder

    >>> grid = RasterModelGrid((5, 5), xy_spacing=10.)
    >>> z = np.array([7.,  7.,  7.,  7.,  7.,
    ...               7.,  5., 3.2,  6.,  7.,
    ...               7.,  2.,  3.,  5.,  7.,
    ...               7.,  1., 1.9,  4.,  7.,
    ...               7.,  0.,  7.,  7.,  7.])
    >>> z = grid.add_field('topographic__elevation', z, at='node')
    >>> fr = FlowAccumulator(grid, flow_director='D8')
    >>> sp = FastscapeEroder(grid, K_sp=1.)
    >>> fr.run_one_step()
    >>> sp.run_one_step(dt=1.)
    >>> z  # doctest: +NORMALIZE_WHITESPACE
    array([ 7.        ,  7.        ,  7.        ,  7.        ,  7.        ,
            7.        ,  2.92996598,  2.02996598,  4.01498299,  7.        ,
            7.        ,  0.85993197,  1.87743897,  3.28268321,  7.        ,
            7.        ,  0.28989795,  0.85403051,  2.42701526,  7.        ,
            7.        ,  0.        ,  7.        ,  7.        ,  7.        ])

    >>> grid = RasterModelGrid((3, 7), xy_spacing=1.)
    >>> z = np.array(grid.node_x ** 2.)
    >>> z = grid.add_field('topographic__elevation', z, at='node')
    >>> grid.status_at_node[grid.nodes_at_left_edge] = grid.BC_NODE_IS_FIXED_VALUE
    >>> grid.status_at_node[grid.nodes_at_top_edge] = grid.BC_NODE_IS_CLOSED
    >>> grid.status_at_node[grid.nodes_at_bottom_edge] = grid.BC_NODE_IS_CLOSED
    >>> grid.status_at_node[grid.nodes_at_right_edge] = grid.BC_NODE_IS_CLOSED
    >>> fr = FlowAccumulator(grid, flow_director='D8')
    >>> sp = FastscapeEroder(grid, K_sp=0.1, m_sp=0., n_sp=2.,
    ...                      threshold_sp=2.)
    >>> fr.run_one_step()
    >>> sp.run_one_step(dt=10.)
    >>> z.reshape(grid.shape)[1, :]  # doctest: +NORMALIZE_WHITESPACE
    array([  0.        ,   1.        ,   4.        ,   8.52493781,
            13.29039716,  18.44367965,  36.        ])

    >>> grid = RasterModelGrid((3, 7), xy_spacing=1.)
    >>> z = np.array(grid.node_x ** 2.)
    >>> z = grid.add_field('topographic__elevation', z, at='node')
    >>> grid.status_at_node[grid.nodes_at_left_edge] = grid.BC_NODE_IS_FIXED_VALUE
    >>> grid.status_at_node[grid.nodes_at_top_edge] = grid.BC_NODE_IS_CLOSED
    >>> grid.status_at_node[grid.nodes_at_bottom_edge] = grid.BC_NODE_IS_CLOSED
    >>> grid.status_at_node[grid.nodes_at_right_edge] = grid.BC_NODE_IS_CLOSED
    >>> cell_area = 1.0
    >>> fr = FlowAccumulator(grid, flow_director='D8', runoff_rate=2.0)
    >>> grid.at_node["water__unit_flux_in"]
    array([ 2.,  2.,  2.,  2.,  2.,  2.,  2.,  2.,  2.,  2.,  2.,  2.,  2.,
            2.,  2.,  2.,  2.,  2.,  2.,  2.,  2.])
    >>> K_field = grid.ones(at='node') # K can be a field
    >>> sp = FastscapeEroder(grid,
    ...     K_sp=K_field,
    ...     m_sp=1.,
    ...     n_sp=0.6,
    ...     threshold_sp=grid.node_x,
    ...     discharge_field="surface_water__discharge")
    >>> fr.run_one_step()
    >>> sp.run_one_step(1.)
    >>> z.reshape(grid.shape)[1, :]  # doctest: +NORMALIZE_WHITESPACE
    array([  0.        ,   0.0647484 ,   0.58634455,   2.67253503,
             8.49212152,  20.92606987,  36.        ])

    References
    ----------
    **Required Software Citation(s) Specific to this Component**

    None Listed

    **Additional References**

    Braun, J., Willett, S. (2013). A very efficient O(n), implicit and parallel
    method to solve the stream power equation governing fluvial incision and
    landscape evolution. Geomorphology  180-181(C), 170-179.
    https://dx.doi.org/10.1016/j.geomorph.2012.10.008

    """

    _name = "FastscapeEroder"

    _unit_agnostic = True

    _info = {
        "drainage_area": {
            "dtype": float,
            "intent": "in",
            "optional": False,
            "units": "m**2",
            "mapping": "node",
            "doc": "Upstream accumulated surface area contributing to the node's discharge",
        },
        "flow__link_to_receiver_node": {
            "dtype": int,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "ID of link downstream of each node, which carries the discharge",
        },
        "flow__receiver_node": {
            "dtype": int,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "Node array of receivers (node that receives flow from current node)",
        },
        "flow__upstream_node_order": {
            "dtype": int,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "Node array containing downstream-to-upstream ordered list of node IDs",
        },
        "topographic__elevation": {
            "dtype": float,
            "intent": "inout",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "Land surface topographic elevation",
        },
    }

    def __init__(
        self,
        grid,
        K_sp=0.001,
        m_sp=0.5,
        n_sp=1.0,
        threshold_sp=0.0,
        discharge_field="drainage_area",
        erode_flooded_nodes=True,
    ):
        """Initialize the Fastscape stream power component. Note: a timestep,
        dt, can no longer be supplied to this component through the input file.
        It must instead be passed directly to the run method.

        Parameters
        ----------
        grid : ModelGrid
            A grid.
        K_sp : float, array, or field name
            K in the stream power equation (units vary with other parameters).
        m_sp : float, optional
            m in the stream power equation (power on drainage area).
        n_sp : float, optional
            n in the stream power equation (power on slope).
        threshold_sp : float, array, or field name
            Erosion threshold in the stream power equation.
        discharge_field : float, field name, or array, optional
            Discharge [L^2/T]. The default is to use the grid field
            'drainage_area'. To use custom spatially/temporally varying
            rainfall, use 'water__unit_flux_in' to specify water input to the
            FlowAccumulator and use "surface_water__discharge" for this
            keyword argument.
        erode_flooded_nodes : bool (optional)
            Whether erosion occurs in flooded nodes identified by a
            depression/lake mapper (e.g., DepressionFinderAndRouter). When set
            to false, the field *flood_status_code* must be present on the grid
            (this is created by the DepressionFinderAndRouter). Default True.
        """
        super().__init__(grid)

        if "flow__receiver_node" in grid.at_node:
            if grid.at_node["flow__receiver_node"].size != grid.size("node"):
                msg = (
                    "A route-to-multiple flow director has been "
                    "run on this grid. The landlab development team has not "
                    "verified that FastscapeEroder is compatible with "
                    "route-to-multiple methods. Please open a GitHub Issue "
                    "to start this process."
                )
                raise NotImplementedError(msg)

        if not erode_flooded_nodes:
            if "flood_status_code" not in self._grid.at_node:
                msg = (
                    "In order to not erode flooded nodes another component "
                    "must create the field *flood_status_code*. You want to "
                    "run a lake mapper/depression finder."
                )
                raise ValueError(msg)

        self._erode_flooded_nodes = erode_flooded_nodes

        # use setter for K defined below
        self.K = K_sp

        self._m = float(m_sp)
        self._n = float(n_sp)

        if isinstance(threshold_sp, (float, int)):
            self._thresholds = float(threshold_sp)
        else:
            self._thresholds = return_array_at_node(grid, threshold_sp)

        self._A = return_array_at_node(grid, discharge_field)

        # make storage variables
        self._A_to_the_m = grid.zeros(at="node")
        self._alpha = grid.empty(at="node")

    @property
    def K(self):
        """Erodibility (units depend on m_sp)."""
        return self._K

    @K.setter
    def K(self, new_val):
        self._K = return_array_at_node(self._grid, new_val)

    def run_one_step(self, dt):
        """Erode for a single time step.

        This method implements the stream power erosion across one time
        interval, dt, following the Braun-Willett (2013) implicit Fastscape
        algorithm.

        This follows Landlab standardized component design, and supercedes the
        old driving method :func:`erode`.

        Parameters
        ----------
        dt : float
            Time-step size
        """
        if not self._erode_flooded_nodes:
            flood_status = self._grid.at_node["flood_status_code"]
            flooded_nodes = np.nonzero(flood_status == _FLOODED)[0]
        else:
            flooded_nodes = []

        upstream_order_IDs = self._grid.at_node["flow__upstream_node_order"]
        flow_receivers = self._grid["node"]["flow__receiver_node"]
        z = self._grid.at_node["topographic__elevation"]

        defined_flow_receivers = np.not_equal(
            self._grid.at_node["flow__link_to_receiver_node"], self._grid.BAD_INDEX
        )

        if isinstance(self._grid, RasterModelGrid):
            flow_link_lengths = self._grid.length_of_d8[
                self._grid.at_node["flow__link_to_receiver_node"][
                    defined_flow_receivers
                ]
            ]
        else:
            flow_link_lengths = self._grid.length_of_link[
                self._grid.at_node["flow__link_to_receiver_node"][
                    defined_flow_receivers
                ]
            ]

        np.power(self._A, self._m, out=self._A_to_the_m)
        self._alpha[defined_flow_receivers] = (
            self._K[defined_flow_receivers]
            * dt
            * self._A_to_the_m[defined_flow_receivers]
            / (flow_link_lengths ** self._n)
        )

        # Handle flooded nodes, if any (no erosion there)
        if len(flooded_nodes) > 0:
            self._alpha[flooded_nodes] = 0.0
        else:
            reversed_flow = z < z[flow_receivers]
            # this check necessary if flow has been routed across depressions
            self._alpha[reversed_flow] = 0.0

        threshsdt = self._thresholds * dt

        # solve using Brent's Method in Cython for Speed
        if isinstance(self._thresholds, float):
            brent_method_erode_fixed_threshold(
                upstream_order_IDs, flow_receivers, threshsdt, self._alpha, self._n, z
            )
        else:
            brent_method_erode_variable_threshold(
                upstream_order_IDs, flow_receivers, threshsdt, self._alpha, self._n, z
            )



# -*- coding: utf-8 -*-
"""
Unit tests for landlab.data_record.data_record.DataRecord
Dimension = time

Last updated 8/24/2018

"""

import numpy as np
import pytest

from landlab import RasterModelGrid

grid = RasterModelGrid((3, 3))
shape = (3, 3)
time = [0.0]
data_vars = {"mean_elevation": (["time"], np.array([100]))}
attrs = {"time_units": "y"}


def test_dr_time_name(dr_time):
    assert dr_time._name == "DataRecord"


def test_grid_shape(dr_time):
    assert dr_time._grid.number_of_node_rows == shape[0]
    assert dr_time._grid.number_of_node_columns == shape[1]


def test_permitted_locations(dr_time):
    assert dr_time._permitted_locations == grid.groups


def test_coordinates(dr_time):
    assert len(dr_time.dataset.dims) == 1
    assert list(dr_time.dataset.time.values) == list(np.array(time))
    assert list(dr_time.time_coordinates) == list(np.array(time))
    # properties:
    assert dr_time.number_of_timesteps == 1
    assert dr_time.earliest_time == 0.0
    assert dr_time.latest_time == 0.0
    assert np.isnan(dr_time.prior_time)
    # no item_id coord:
    with pytest.raises(AttributeError):
        dr_time.item_id
    with pytest.raises(AttributeError):
        dr_time.item_coordinates
    with pytest.raises(AttributeError):
        dr_time.number_of_items


def test_variable_names(dr_time):
    assert dr_time.variable_names == ["mean_elevation"]


def test_add_record(dr_time):
    dr_time.add_record(
        time=[50.0], new_record={"mean_elevation": (["time"], np.array([120]))}
    )
    dr_time.add_record(
        time=[100.0], new_record={"new_variable": (["time"], ["new_data"])}
    )
    assert np.isnan(dr_time.dataset["mean_elevation"].values[2])


def test_get_data(dr_time):
    assert dr_time.get_data(time=[0.0], data_variable="mean_elevation") == 100.0
    assert dr_time.get_data(data_variable="mean_elevation") == [100]


def test_set_data(dr_time):
    dr_time.set_data(time=[0.0], data_variable="mean_elevation", new_value=105.0)
    assert dr_time.dataset["mean_elevation"].values[0] == 105.0



import numpy as np
from numpy.testing import assert_array_equal

from landlab.graph.sort.ext.remap_element import offset_to_sorted_block
from landlab.graph.sort.intpair import (
    map_pairs_to_values,
    map_rolling_pairs_to_values,
    pair_isin,
)


def test_pair_isin_one_pair():
    src = np.asarray(
        [[0, 1], [1, 2], [2, 5], [5, 8], [8, 7], [7, 6], [6, 3], [3, 0]], dtype=int
    )
    pairs = np.asarray([[4, 8]], dtype=int)
    out = pair_isin(src, pairs)

    assert tuple(out) == tuple([False])


def test_pair_isin_1():
    src = np.asarray(
        [[0, 1], [1, 2], [2, 5], [5, 8], [8, 7], [7, 6], [6, 3], [3, 0]], dtype=int
    )
    pairs = np.asarray(
        [
            [5, 2],
            [5, 4],
            [5, 8],
            [1, 2],
            [1, 4],
            [1, 0],
            [1, 3],
            [2, 4],
            [0, 6],
            [0, 3],
            [7, 3],
            [7, 4],
            [7, 6],
            [7, 8],
            [3, 6],
            [3, 4],
            [4, 8],
        ],
        dtype=int,
    )
    out = pair_isin(src, pairs, sorter=np.argsort(src[:, 0]))

    assert tuple(out) == tuple(
        [
            True,
            False,
            True,
            True,
            False,
            True,
            False,
            False,
            False,
            True,
            False,
            False,
            True,
            True,
            True,
            False,
            False,
        ]
    )


def test_pair_isin():
    src = np.asarray([[0, 1], [1, 1], [2, 1], [3, 1], [4, 1]], dtype=int)

    pairs = np.asarray([[1, 1], [3, 1], [1, 2], [5, 1]], dtype=int)
    out = pair_isin(src, pairs)

    assert np.all(out == [True, True, True, False])


def test_pair_isin_with_out_keyword():
    src = np.asarray([[0, 1], [1, 1], [2, 1], [3, 1], [4, 1]], dtype=int)

    pairs = np.asarray([[1, 1], [3, 1], [1, 2], [5, 1]], dtype=int)
    out = np.empty(len(pairs), dtype=bool)
    rtn = pair_isin(src, pairs, out=out)

    assert rtn is out
    assert np.all(out == [True, True, True, False])


def test_map_pairs():
    src = np.asarray([[0, 1], [1, 1], [2, 1], [3, 1], [4, 1]], dtype=int)
    data = np.arange(len(src), dtype=int)

    pairs = np.asarray([[1, 1], [3, 1]], dtype=int)
    out = map_pairs_to_values((src, data), pairs)

    assert np.all(out == [1, 3])


def test_map_pairs_with_out_keyword():
    src = np.asarray([[0, 1], [1, 1], [2, 1], [3, 1], [4, 1]], dtype=int)
    data = np.arange(len(src), dtype=int)

    pairs = np.asarray([[1, 1], [3, 1]], dtype=int)
    out = np.empty(len(pairs), dtype=int)
    rtn = map_pairs_to_values((src, data), pairs, out=out)

    assert rtn is out
    assert np.all(out == [1, 3])


def test_map_pairs_with_sorter_keyword():
    src = np.asarray([[0, 1], [1, 1], [3, 1], [2, 1], [4, 1]], dtype=int)
    data = np.array([0, 1, 3, 2, 4], dtype=int)

    pairs = np.asarray([[1, 1], [3, 1]], dtype=int)
    out = map_pairs_to_values((src, data), pairs, sorter=np.argsort(src[:, 0]))

    assert np.all(out == [1, 3])


def test_map_pairs_at_end():
    src = np.asarray([[0, 1], [1, 1], [2, 1], [3, 1], [4, 1]], dtype=int)
    data = np.arange(len(src), dtype=int)

    pairs = np.asarray([[1, 1], [3, 1], [4, 1], [1, 4], [4, 2]], dtype=int)
    out = map_pairs_to_values((src, data), pairs)

    assert np.all(out == [1, 3, 4, 4, -1])


def test_map_pairs_transposed():
    src = np.asarray([[0, 1], [1, 1], [2, 1], [3, 1], [4, 1]], dtype=int)
    data = np.arange(len(src), dtype=int)

    pairs = np.asarray([[1, 0], [1, 3]], dtype=int)
    out = map_pairs_to_values((src, data), pairs)

    assert np.all(out == [0, 3])


def test_map_pairs_all_missing():
    src = np.asarray([[0, 1], [1, 1], [2, 1], [3, 1], [4, 1]], dtype=int)
    data = np.arange(len(src), dtype=int)

    pairs = np.asarray([[5, 1], [1, 42]], dtype=int)
    out = map_pairs_to_values((src, data), pairs)

    assert np.all(out == [-1, -1])


def test_map_pairs_big_data():
    n_src_pairs = 2 ** 20
    src = np.random.randint(n_src_pairs * 2, size=n_src_pairs * 2, dtype=int).reshape(
        (-1, 2)
    )
    src = src[np.lexsort((src[:, 1], src[:, 0]))]
    data = np.arange(n_src_pairs, dtype=int)

    ids = np.random.randint(n_src_pairs, size=n_src_pairs // 10)
    pairs = src[ids]
    out = map_pairs_to_values((src, data), pairs)

    assert_array_equal(out, ids)


def test_map_rolling_pairs():
    src = np.asarray([[0, 1], [1, 2], [2, 3], [3, 4], [4, 0]], dtype=int)
    data = np.asarray([0, 1, 2, 3, 4], dtype=int)

    pairs = np.asarray([[0, 1, 2, 3], [0, 2, 3, 4]], dtype=int)
    out = map_rolling_pairs_to_values((src, data), pairs)

    assert np.all(out == [[0, 1, 2, -1], [-1, 2, 3, 4]])


def test_map_rolling_pairs_with_out_keyword():
    src = np.asarray([[0, 1], [1, 2], [2, 3], [3, 4], [4, 0]], dtype=int)
    data = np.asarray([0, 1, 2, 3, 4], dtype=int)

    pairs = np.asarray([[0, 1, 2, 3], [0, 2, 3, 4]], dtype=int)
    out = np.empty_like(pairs)
    rtn = map_rolling_pairs_to_values((src, data), pairs, out=out)

    assert rtn is out
    assert np.all(out == [[0, 1, 2, -1], [-1, 2, 3, 4]])


def test_map_rolling_pairs_with_size_of_row():
    src = np.asarray([[0, 1], [1, 2], [2, 3], [2, 0], [3, 4], [4, 0]], dtype=int)
    data = np.asarray([0, 1, 2, 3, 4, 5], dtype=int)

    pairs = np.asarray([[0, 1, 2, -1], [0, 2, -1, -1]], dtype=int)
    out = map_rolling_pairs_to_values((src, data), pairs, size_of_row=[3, 2])

    assert np.all(out == [[0, 1, 3, -1], [3, 3, -1, -1]])


def test_offset_2d():
    pairs = np.asarray([[0, 1], [1, 1], [2, 1], [3, 1], [4, 1]], dtype=int)
    out = np.empty(len(pairs), dtype=int)
    offset_to_sorted_block(pairs, out)

    assert np.all(out == [0, 1, 2, 3, 4])


def test_offset_2d_with_missing_at_start():
    pairs = np.asarray([[1, 1], [2, 1], [3, 1], [4, 1]], dtype=int)
    out = np.empty(5, dtype=int)
    offset_to_sorted_block(pairs, out)

    assert np.all(out == [0, 0, 1, 2, 3])


def test_offset_2d_with_missing_in_middle():
    pairs = np.asarray([[0, 1], [1, 1], [3, 1], [4, 1]], dtype=int)
    out = np.empty(5, dtype=int)
    offset_to_sorted_block(pairs, out)

    assert np.all(out == [0, 1, 2, 2, 3])


def test_offset_with_different_strides():
    pairs = np.arange(5, dtype=int).reshape((-1, 1))
    out = np.empty(len(pairs), dtype=int)
    for _ in range(5):
        offset_to_sorted_block(pairs, out)
        assert np.all(out == [0, 1, 2, 3, 4])
        pairs = np.hstack((pairs, pairs))


def test_offset_with_sort_out():
    pairs = np.asarray([[1, 1], [2, 1], [3, 1], [4, 1]], dtype=int)
    out = np.empty(4, dtype=int)
    offset_to_sorted_block(pairs, out)

    assert np.all(out == [0, 0, 1, 2])


def test_offset_with_long_out():
    pairs = np.asarray([[1, 1], [2, 1], [3, 1], [4, 1]], dtype=int)
    out = np.empty(7, dtype=int)
    offset_to_sorted_block(pairs, out)

    assert np.all(out == [0, 0, 1, 2, 3, 4, 4])


def test_offset_to_sorted_block_with_negatives():
    pairs = np.array(
        [
            [-1, 0],
            [-1, 8],
            [-1, 7],
            [0, 3],
            [0, 2],
            [1, 4],
            [1, 7],
            [2, 8],
            [3, 5],
            [4, 5],
            [5, 6],
            [6, 7],
            [6, 8],
        ]
    )
    out = np.empty(len(pairs), dtype=int)
    offset_to_sorted_block(pairs, out)
    assert tuple(out) == (3, 5, 7, 8, 9, 10, 11, 13, 13, 13, 13, 13, 13)


def test_offset_to_sorted_block_all_negatives():
    pairs = np.array([[-1, 0], [-1, 8], [-1, 7]])
    out = np.empty(len(pairs) + 1, dtype=int)
    offset_to_sorted_block(pairs, out)
    assert tuple(out) == (3, 3, 3, 3)



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Thu Jul 27 14:23:25 2017

@author: gtucker
"""

import numpy as np
from numpy.testing import assert_equal

from landlab import RasterModelGrid
from landlab.components import ErosionDeposition, FlowAccumulator


def test_erodep_slope_area_small_vs():
    """Test steady state run with Vs << 1."""

    # Set up a 5x5 grid with open boundaries and low initial elevations.
    rg = RasterModelGrid((5, 5))
    z = rg.add_zeros("topographic__elevation", at="node")
    z[:] = 0.01 * rg.x_of_node

    # Create a D8 flow handler
    fa = FlowAccumulator(rg, flow_director="FlowDirectorD8")

    # Parameter values for test 1
    K = 0.001
    vs = 0.0001
    U = 0.001
    dt = 10.0

    # Create the ErosionDeposition component...
    ed = ErosionDeposition(rg, K=K, v_s=vs, m_sp=0.5, n_sp=1.0, solver="adaptive")

    # ... and run it to steady state.
    for i in range(1000):
        fa.run_one_step()
        ed.run_one_step(dt=dt)
        z[rg.core_nodes] += U * dt

    # Test the results
    s = rg.at_node["topographic__steepest_slope"]
    sa_factor = (1.0 + vs) * U / K
    a11 = 2.0
    a12 = 1.0
    s = rg.at_node["topographic__steepest_slope"]
    s11 = sa_factor * (a11 ** -0.5)
    s12 = sa_factor * (a12 ** -0.5)
    assert_equal(np.round(s[11], 3), np.round(s11, 3))
    assert_equal(np.round(s[12], 3), np.round(s12, 3))


def test_erodep_slope_area_big_vs():
    """Test steady state run with Vs >> 1."""

    # Set up a 5x5 grid with open boundaries and low initial elevations.
    rg = RasterModelGrid((5, 5))
    z = rg.add_zeros("topographic__elevation", at="node")
    z[:] = 0.01 * rg.x_of_node

    # Create a D8 flow handler
    fa = FlowAccumulator(rg, flow_director="FlowDirectorD8")

    # Next test: big Vs
    K = 1.0
    vs = 1000.0
    U = 0.001
    dt = 10.0

    # Create the ErosionDeposition component...
    ed = ErosionDeposition(rg, K=K, v_s=vs, m_sp=0.5, n_sp=1.0, solver="adaptive")

    # ... and run it to steady state.
    for i in range(1000):
        fa.run_one_step()
        ed.run_one_step(dt=dt)
        z[rg.core_nodes] += U * dt

    # Test the results
    s = rg.at_node["topographic__steepest_slope"]
    sa_factor = (1.0 + vs) * U / K
    a11 = 2.0
    a12 = 1.0
    s11 = sa_factor * (a11 ** -0.5)
    s12 = sa_factor * (a12 ** -0.5)
    assert_equal(np.round(s[11], 2), np.round(s11, 2))
    assert_equal(np.round(s[12], 2), np.round(s12, 2))


def test_erodep_slope_area_with_vs_unity():
    """Test steady state run with Vs = 1."""

    # Set up a 5x5 grid with open boundaries and low initial elevations.
    rg = RasterModelGrid((5, 5))
    z = rg.add_zeros("topographic__elevation", at="node")
    z[:] = 0.01 * rg.x_of_node

    # Create a D8 flow handler
    fa = FlowAccumulator(rg, flow_director="FlowDirectorD8")

    # test: Vs = 1
    K = 0.002
    vs = 1.0
    U = 0.001
    dt = 10.0

    # Create the ErosionDeposition component...
    ed = ErosionDeposition(rg, K=K, v_s=vs, m_sp=0.5, n_sp=1.0, solver="adaptive")

    # ... and run it to steady state.
    for i in range(1000):
        fa.run_one_step()
        ed.run_one_step(dt=dt)
        z[rg.core_nodes] += U * dt

    # Test the results
    s = rg.at_node["topographic__steepest_slope"]
    sa_factor = (1.0 + vs) * U / K
    a11 = 2.0
    a12 = 1.0
    s11 = sa_factor * (a11 ** -0.5)
    s12 = sa_factor * (a12 ** -0.5)
    assert_equal(np.round(s[11], 2), np.round(s11, 2))
    assert_equal(np.round(s[12], 2), np.round(s12, 2))


def test_erodep_slope_area_shear_stress_scaling():
    """Test steady state run with m_sp = 0.33, n_sp=0.67, Vs = 1."""

    # Set up a 5x5 grid with open boundaries and low initial elevations.
    rg = RasterModelGrid((5, 5))
    rg.set_closed_boundaries_at_grid_edges(True, True, True, False)
    z = rg.add_zeros("topographic__elevation", at="node")
    z[:] = 0.01 * rg.x_of_node

    # Create a D8 flow handler
    fa = FlowAccumulator(rg, flow_director="FlowDirectorD8")

    # test: Vs = 1
    K = 0.002
    vs = 1.0
    U = 0.001
    dt = 10.0
    m_sp = 0.33
    n_sp = 0.67
    # Create the ErosionDeposition component...
    ed = ErosionDeposition(rg, K=K, v_s=vs, m_sp=m_sp, n_sp=n_sp, solver="adaptive")

    # ... and run it to steady state.
    for i in range(1500):
        fa.run_one_step()
        ed.run_one_step(dt=dt)
        z[rg.core_nodes] += U * dt

    # Test the results
    s = rg.at_node["topographic__steepest_slope"]
    sa_factor = ((1.0 + vs) * U / K) ** (1.0 / n_sp)
    a6 = rg.at_node["drainage_area"][6]
    a8 = rg.at_node["drainage_area"][8]
    s6 = sa_factor * (a6 ** -(m_sp / n_sp))
    s8 = sa_factor * (a8 ** -(m_sp / n_sp))
    assert_equal(np.round(s[6], 2), np.round(s6, 2))
    assert_equal(np.round(s[8], 2), np.round(s8, 2))


def test_erodep_slope_area_with_threshold():
    """Test steady state run with Vs = 1 and wc = 0.00001."""

    # Set up a 5x5 grid with open boundaries and low initial elevations.
    rg = RasterModelGrid((5, 5))
    z = rg.add_zeros("topographic__elevation", at="node")
    z[:] = 0.01 * rg.x_of_node

    # Create a D8 flow handler
    fa = FlowAccumulator(rg, flow_director="FlowDirectorD8")

    # test: Vs = 1
    K = 0.002
    vs = 1.0
    U = 0.001
    dt = 10.0
    wc = 0.0001

    # Create the ErosionDeposition component...
    ed = ErosionDeposition(
        rg, K=K, v_s=vs, m_sp=0.5, n_sp=1.0, sp_crit=wc, solver="adaptive"
    )

    # ... and run it to steady state.
    for i in range(1000):
        fa.run_one_step()
        ed.run_one_step(dt=dt)
        z[rg.core_nodes] += U * dt

    # Test the results
    s = rg.at_node["topographic__steepest_slope"]
    sa_factor = ((1.0 + vs) * U + wc) / K  # approximate sol'n
    a11 = 2.0
    a12 = 1.0
    s11 = sa_factor * (a11 ** -0.5)
    s12 = sa_factor * (a12 ** -0.5)
    assert_equal(np.round(s[11], 2), np.round(s11, 2))
    assert_equal(np.round(s[12], 2), np.round(s12, 2))



from landlab.plot.imshow import (
    imshow_grid,
    imshow_grid_at_node,
    imshowhs_grid,
    imshowhs_grid_at_node,
)
from landlab.plot.network_sediment_transporter import plot_network_and_parcels

__all__ = [
    "imshow_grid",
    "imshowhs_grid",
    "imshow_grid_at_node",
    "imshowhs_grid_at_node",
    "plot_network_and_parcels",
]



#! /usr/env/python

"""Create 2D grid with randomly generated fractures.

Created: September 2013 by Greg Tucker
Last significant modification: conversion to proper component 7/2019 GT
"""

import numpy as np

from landlab import Component, HexModelGrid, RasterModelGrid


def _calc_fracture_starting_position_raster(shape):
    """Choose a random starting position along one of the sides of the grid.

    Parameters
    ----------
    shape : tuple of int
        Number of rows and columns in the grid

    Returns
    -------
    (c, r) : tuple of int
        Fracture starting coordinates (column and row IDs)
    """
    grid_side = np.random.randint(0, 3)  # east, north, west, south

    if (grid_side % 2) == 0:  # east or west
        c = (1 - grid_side // 2) * (shape[1] - 1)
        r = np.random.randint(0, shape[0] - 1)
    else:
        c = np.random.randint(0, shape[1] - 1)
        r = (1 - grid_side // 2) * (shape[0] - 1)

    return (c, r)


def _calc_fracture_starting_position_and_angle_hex(shape, is_horiz, spacing):
    """Choose a random starting position along one of the sides of the grid.

    Parameters
    ----------
    shape : tuple of int
        Number of rows and columns in the grid

    Returns
    -------
    (x, y, ang) : tuple of float
        Fracture starting coordinates and angle (radians)
    """
    grid_side = np.random.randint(0, 3)  # east, north, west, south
    ang = np.pi * np.random.rand()
    if is_horiz:
        row_spacing = 0.5 * 3.0 ** 0.5 * spacing
        col_spacing = spacing
    else:
        row_spacing = spacing
        col_spacing = 0.5 * 3.0 ** 0.5 * spacing

    if (grid_side % 2) == 0:  # east or west
        c = (1 - grid_side // 2) * (shape[1] - 1)
        r = np.random.randint(0, shape[0] - 1)
        if grid_side == 0:  # east
            ang += np.pi / 2
        else:  # west
            ang += 1.5 * np.pi
    else:
        c = np.random.randint(0, shape[1] - 1)
        r = (1 - grid_side // 2) * (shape[0] - 1)
        if grid_side == 1:  # north
            ang += np.pi

    x = c * col_spacing
    y = r * row_spacing

    epsilon = 0.001 * spacing  # tiny offset to ensure points start inside grid
    if x > epsilon:
        x -= epsilon
    if y > epsilon:
        y -= epsilon

    return (x, y, ang)


def _calc_fracture_orientation(coords, shape):
    """Choose a random orientation for the fracture.

    Parameters
    ----------
    coords : tuple of int
        Starting coordinates (one of which should be zero) as *x*, *y*.
    shape : tuple of int
        Number of rows and columns

    Returns
    -------
    ang : float
        Fracture angle relative to horizontal

    Notes
    -----
    The angle depends on which side of the grid the fracture starts on:
        - east: 90-270
        - north: 180-360
        - west: 270-450 (mod 360)
        - south: 0-180
    """
    x, y = coords
    ang = np.pi * np.random.rand()

    if x == shape[1] - 1:  # east
        ang += np.pi / 2
    elif y == shape[0] - 1:  # north
        ang += np.pi
    elif x == 0:  # west
        ang += 1.5 * np.pi

    return ang


def _calc_fracture_step_sizes(ang):
    """Calculate the sizes of steps dx and dy to be used when "drawing" the
    fracture onto the grid.

    Parameters
    ----------
    start_yx : tuple of int
        Starting grid coordinates
    ang : float
        Fracture angle relative to horizontal (radians)

    Returns
    -------
    (dy, dx) : tuple of float
        Step sizes in y and x directions. One will always be unity, and the
        other will always be <1.

    Examples
    --------
    >>> np.round(_calc_fracture_step_sizes(0 * np.pi / 6), 3)
    array([ 1.,  0.])
    >>> np.round(_calc_fracture_step_sizes(1 * np.pi / 6), 3)
    array([ 1.   , 0.577])
    >>> np.round(_calc_fracture_step_sizes(2 * np.pi / 6), 3)
    array([ 0.577,  1.   ])
    >>> np.round(_calc_fracture_step_sizes(3 * np.pi / 6), 3)
    array([ 0., 1.])
    >>> np.round(_calc_fracture_step_sizes(4 * np.pi / 6), 3)
    array([-0.577, 1.   ])
    >>> np.round(_calc_fracture_step_sizes(5 * np.pi / 6), 3)
    array([-1.   , 0.577])
    >>> np.round(_calc_fracture_step_sizes(6 * np.pi / 6), 3)
    array([-1., 0.])
    >>> np.round(_calc_fracture_step_sizes(7 * np.pi / 6), 3)
    array([-1.   , -0.577])
    >>> np.round(_calc_fracture_step_sizes(8 * np.pi / 6), 3)
    array([-0.577, -1.   ])
    >>> np.round(_calc_fracture_step_sizes(9 * np.pi / 6), 3)
    array([-0., -1.])
    >>> np.round(_calc_fracture_step_sizes(10 * np.pi / 6), 3)
    array([ 0.577, -1.   ])
    >>> np.round(_calc_fracture_step_sizes(11 * np.pi / 6), 3)
    array([ 1.   , -0.577])
    >>> np.round(_calc_fracture_step_sizes(12 * np.pi / 6), 3)
    array([ 1., -0.])
    """
    dx = np.cos(ang)
    dy = np.sin(ang)
    multiplier = 1.0 / max(np.abs(dx), np.abs(dy))
    dx *= multiplier
    dy *= multiplier

    return dx, dy


def _trace_fracture_through_grid_raster(m, start_xy, spacing):
    """Create a 2D fracture in a grid.

    Creates a "fracture" in a 2D grid, m, by setting cell values to unity
    along the trace of the fracture (i.e., "drawing" a line throuh the
    grid).

    Parameters
    ----------
    m : 2D Numpy array
        Array that represents the grid
    start_xy : tuple of int
        Starting grid coordinates (col, row) for fracture
    spacing : tuple of float
        Step sizes in x and y directions

    Returns
    -------
    None, but changes contents of m
    """
    x0, y0 = start_xy
    dx, dy = spacing

    x = x0
    y = y0

    while (
        round(x) < np.size(m, 1)
        and round(y) < np.size(m, 0)
        and round(x) >= 0
        and round(y) >= 0
    ):
        m[int(y + 0.5)][int(x + 0.5)] = 1
        x += dx
        y += dy


class FractureGridGenerator(Component):

    """Create a 2D grid with randomly generated fractures.

    The grid contains the value 1 where fractures (one cell wide) exist, and
    0 elsewhere. The idea is to use this for simulations based on weathering
    and erosion of, and/or flow within, fracture networks.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> grid = RasterModelGrid((5, 5))
    >>> fg = FractureGridGenerator(grid=grid, frac_spacing=3)
    >>> fg.run_one_step()
    >>> grid.at_node['fracture_at_node'].reshape((5, 5))
    array([[1, 0, 0, 1, 0],
           [0, 1, 1, 1, 1],
           [0, 0, 0, 1, 1],
           [0, 0, 0, 1, 1],
           [0, 0, 0, 1, 0]], dtype=int8)

    Notes
    -----
    Potential improvements:

    - Fractures could be defined by links rather than nodes (i.e., return a
        link array with a code indicating whether the link crosses a fracture
        or not)
    - Fractures could have a finite length rather than extending all the way
        across the grid
    - Use of starting position along either x or y axis makes fracture net
        somewhat asymmetric. One would need a different algorithm to make it
        fully (statistically) symmetric.

    References
    ----------
    **Required Software Citation(s) Specific to this Component**

    None Listed

    **Additional References**

    None Listed

    """

    _name = "FractureGridGenerator"

    _unit_agnostic = True

    _info = {
        "fracture_at_node": {
            "dtype": np.int8,
            "intent": "out",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "presence (1) or absence (0) of fracture",
        }
    }

    def __init__(self, grid, frac_spacing=10.0, seed=0):
        """Initialize the FractureGridGenerator.

        Parameters
        ----------
        frac_spacing : int, optional
            Average spacing of fractures (in grid cells) (default = 10)
        seed : int, optional
            Seed used for random number generator (default = 0)

        """

        self._frac_spacing = frac_spacing
        self._seed = seed
        super().__init__(grid)

        if isinstance(grid, RasterModelGrid):
            self._make_frac_grid = self.make_frac_grid_raster
        elif isinstance(grid, HexModelGrid):
            self._make_frac_grid = self.make_frac_grid_hex
        else:
            raise TypeError("grid must be RasterModelGrid or HexModelGrid")

        if "fracture_at_node" not in grid.at_node:
            grid.add_zeros("node", "fracture_at_node", dtype=np.int8)

        np.random.seed(seed)

    def run_one_step(self):
        """Run FractureGridGenerator and create a random fracture grid."""
        self._make_frac_grid(self._frac_spacing)

    def make_frac_grid_raster(self, frac_spacing):
        """Create a raster grid that contains a network of random fractures.

        Creates a grid containing a network of random fractures, which are
        represented as 1's embedded in a grid of 0's. The grid is stored in
        the "fracture_at_node" field.

        Parameters
        ----------
        frac_spacing : int
            Average spacing of fractures (in grid cells)
        """
        # Make an initial grid of all zeros. If user specified a model grid,
        # use that. Otherwise, use the given dimensions.
        nr = self._grid.number_of_node_rows
        nc = self._grid.number_of_node_columns
        m = self._grid.at_node["fracture_at_node"].reshape((nr, nc))

        # Add fractures to grid
        nfracs = (nr + nc) // frac_spacing
        for i in range(nfracs):

            (c, r) = _calc_fracture_starting_position_raster((nr, nc))
            ang = _calc_fracture_orientation((c, r), (nr, nc))
            (dx, dy) = _calc_fracture_step_sizes(ang)

            _trace_fracture_through_grid_raster(m, (c, r), (dx, dy))

    def make_frac_grid_hex(self, frac_spacing):
        """Create a hex grid that contains a network of random fractures.

        Creates a grid containing a network of random fractures, which are
        represented as 1's embedded in a grid of 0's. The grid is stored in
        the "fracture_at_node" field.

        Parameters
        ----------
        frac_spacing : int
            Average spacing of fractures (in # of grid-cell widths)
        """
        # Make an initial grid of all zeros
        nr = self._grid.number_of_node_rows
        nc = self._grid.number_of_node_columns
        m = self._grid.at_node["fracture_at_node"]  # .reshape((nr, nc))

        # Add fractures to grid
        nfracs = (nr + nc) // frac_spacing
        for i in range(nfracs):

            (x, y, ang) = _calc_fracture_starting_position_and_angle_hex(
                (nr, nc),
                is_horiz=(self._grid.orientation[0] == "h"),
                spacing=self._grid.spacing,
            )

            dx = 0.5 * np.cos(ang)
            dy = 0.5 * np.sin(ang)

            # the following is a TERRIBLE brute-force, algorithm, with its only
            # redeeming feature being that it was quick to code
            xmax = np.amax(self._grid.x_of_node)
            ymax = np.amax(self._grid.y_of_node)
            while x >= 0 and x <= xmax and y >= 0 and y <= ymax:
                distx2 = (self._grid.x_of_node - x) ** 2
                disty2 = (self._grid.y_of_node - y) ** 2
                dist = np.sqrt(distx2 + disty2)
                closest_node = np.argmin(dist)
                x += dx
                y += dy
                m[closest_node] = 1



# coding: utf8
# ! /usr/env/python
"""
Created on Tue Feb 27 16:25:11 2018

@author: barnhark
"""
import matplotlib
import numpy as np
import pytest

from landlab import FieldError, RasterModelGrid
from landlab.components import (
    ChannelProfiler,
    DepressionFinderAndRouter,
    FastscapeEroder,
    FlowAccumulator,
    LinearDiffuser,
)

matplotlib.use("agg")


def test_assertion_error():
    """Test that the correct assertion error will be raised."""
    mg = RasterModelGrid((10, 10))
    z = mg.add_zeros("topographic__elevation", at="node")
    z += 200 + mg.x_of_node + mg.y_of_node + np.random.randn(mg.size("node"))

    mg.set_closed_boundaries_at_grid_edges(
        bottom_is_closed=True,
        left_is_closed=True,
        right_is_closed=True,
        top_is_closed=True,
    )
    mg.set_watershed_boundary_condition_outlet_id(0, z, -9999)
    fa = FlowAccumulator(
        mg, flow_director="D8", depression_finder=DepressionFinderAndRouter
    )
    sp = FastscapeEroder(mg, K_sp=0.0001, m_sp=0.5, n_sp=1, erode_flooded_nodes=True)
    ld = LinearDiffuser(mg, linear_diffusivity=0.0001)

    dt = 100
    for i in range(200):
        fa.run_one_step()
        sp.run_one_step(dt=dt)
        ld.run_one_step(dt=dt)
        mg.at_node["topographic__elevation"][0] -= 0.001  # Uplift

    with pytest.raises(ValueError):
        ChannelProfiler(mg, outlet_nodes=[0], number_of_watersheds=2)


def test_asking_for_too_many_watersheds():
    mg = RasterModelGrid((10, 10))
    z = mg.add_zeros("topographic__elevation", at="node")
    z += 200 + mg.x_of_node + mg.y_of_node
    mg.set_closed_boundaries_at_grid_edges(
        bottom_is_closed=True,
        left_is_closed=True,
        right_is_closed=True,
        top_is_closed=True,
    )
    mg.set_watershed_boundary_condition_outlet_id(0, z, -9999)
    fa = FlowAccumulator(mg, flow_director="D8")
    sp = FastscapeEroder(mg, K_sp=0.0001, m_sp=0.5, n_sp=1)

    dt = 100
    for i in range(200):
        fa.run_one_step()
        sp.run_one_step(dt=dt)
        mg.at_node["topographic__elevation"][0] -= 0.001

    with pytest.raises(ValueError):
        ChannelProfiler(mg, number_of_watersheds=3)

    with pytest.raises(ValueError):
        ChannelProfiler(mg, number_of_watersheds=None, minimum_outlet_threshold=200)


def test_no_minimum_channel_threshold():
    mg = RasterModelGrid((10, 10))
    z = mg.add_zeros("topographic__elevation", at="node")
    z += 200 + mg.x_of_node + mg.y_of_node + np.random.randn(mg.size("node"))

    mg.set_closed_boundaries_at_grid_edges(
        bottom_is_closed=True,
        left_is_closed=True,
        right_is_closed=True,
        top_is_closed=True,
    )
    mg.set_watershed_boundary_condition_outlet_id(0, z, -9999)
    fa = FlowAccumulator(
        mg, flow_director="D8", depression_finder=DepressionFinderAndRouter
    )
    fa.run_one_step()

    profiler = ChannelProfiler(mg)

    assert profiler._minimum_channel_threshold == 0.0


def test_no_flow__link_to_receiver_node():
    mg = RasterModelGrid((10, 10))
    mg.add_zeros("topographic__elevation", at="node")
    mg.add_zeros("drainage_area", at="node")
    mg.add_zeros("flow__receiver_node", at="node")
    with pytest.raises(FieldError):
        ChannelProfiler(mg)


def test_no_flow__receiver_node():
    mg = RasterModelGrid((10, 10))
    mg.add_zeros("topographic__elevation", at="node")
    mg.add_zeros("drainage_area", at="node")
    mg.add_zeros("flow__link_to_receiver_node", at="node")
    with pytest.raises(FieldError):
        ChannelProfiler(mg)


@pytest.fixture()
def profile_example_grid():
    mg = RasterModelGrid((40, 60))
    z = mg.add_zeros("topographic__elevation", at="node")
    z += 200 + mg.x_of_node + mg.y_of_node
    mg.set_closed_boundaries_at_grid_edges(
        bottom_is_closed=True,
        left_is_closed=True,
        right_is_closed=True,
        top_is_closed=True,
    )
    mg.set_watershed_boundary_condition_outlet_id(0, z, -9999)
    fa = FlowAccumulator(mg, flow_director="D8")
    sp = FastscapeEroder(mg, K_sp=0.0001, m_sp=0.5, n_sp=1)

    dt = 100
    for i in range(200):
        fa.run_one_step()
        sp.run_one_step(dt=dt)
        mg.at_node["topographic__elevation"][0] -= 0.001
    return mg


def test_start_away_from_boundary(profile_example_grid):
    mg = profile_example_grid
    profiler = ChannelProfiler(mg, outlet_nodes=[64], cmap="viridis_r")
    profiler.run_one_step()
    assert profiler.distance_along_profile[0][0] == 0.0


def test_plotting_and_structure(profile_example_grid):
    mg = profile_example_grid
    profiler = ChannelProfiler(
        mg,
        number_of_watersheds=1,
        main_channel_only=False,
        minimum_channel_threshold=50,
    )
    profiler.run_one_step()

    profiler.plot_profiles()
    profiler.plot_profiles_in_map_view()

    # hard to test plotting... but in April 2019 KRB visually verified that the
    # plots were correct and has hard coded in what the profile structure was.
    correct_structure = np.array(
        [
            np.array([0, 61]),
            np.array([61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71]),
            np.array([61, 121, 181, 241, 301, 361, 421, 481, 541, 601, 661, 721, 781]),
            np.array([71, 72, 73, 74, 75]),
            np.array([71, 131, 191, 251, 311, 371, 431, 491, 551]),
            np.array([781, 841, 901, 961, 1021]),
            np.array([781, 842, 843, 844, 845, 846, 847]),
            np.array([75, 76, 77, 78, 79]),
            np.array(
                [
                    75,
                    135,
                    195,
                    255,
                    315,
                    375,
                    435,
                    495,
                    555,
                    615,
                    675,
                    735,
                    795,
                    855,
                    915,
                    975,
                    1035,
                ]
            ),
            np.array([1021, 1081, 1141, 1201, 1261, 1321]),
            np.array([1021, 1082, 1083, 1084, 1085, 1086, 1087, 1088]),
            np.array([79, 80, 81, 82, 83]),
            np.array(
                [
                    79,
                    139,
                    199,
                    259,
                    319,
                    379,
                    439,
                    499,
                    559,
                    619,
                    679,
                    739,
                    799,
                    859,
                    919,
                    979,
                    1039,
                    1099,
                ]
            ),
            np.array([1321, 1322, 1323, 1324]),
            np.array([1321, 1381, 1441, 1501, 1561, 1621, 1681, 1741, 1801]),
            np.array([83, 84, 85, 86]),
            np.array(
                [
                    83,
                    143,
                    203,
                    263,
                    323,
                    383,
                    443,
                    503,
                    563,
                    623,
                    683,
                    743,
                    803,
                    863,
                    923,
                    983,
                ]
            ),
            np.array([86, 87, 88, 89, 90]),
            np.array(
                [
                    86,
                    147,
                    207,
                    267,
                    327,
                    387,
                    447,
                    507,
                    567,
                    627,
                    687,
                    747,
                    807,
                    867,
                    927,
                    987,
                    1047,
                ]
            ),
            np.array([90, 91, 92, 93, 94, 95]),
            np.array([90, 151, 211, 271, 331, 391, 451, 511, 571, 631, 691, 751]),
            np.array([95, 96, 97, 98]),
            np.array([95, 155, 215, 275, 335, 395, 455, 515, 575, 635]),
            np.array([98, 99, 100, 101, 102, 103]),
            np.array([98, 159, 219, 279, 339, 399, 459]),
            np.array([103, 104, 105, 106, 107, 108, 109]),
            np.array([103, 163]),
        ]
    )
    for idx in range(len(correct_structure)):
        np.testing.assert_array_equal(profiler.nodes[idx], correct_structure[idx])


def test_end_nodes_only(profile_example_grid):
    mg = profile_example_grid
    # with the same grid, test some other profiler options.
    profiler2 = ChannelProfiler(
        mg,
        number_of_watersheds=None,
        main_channel_only=True,
        minimum_outlet_threshold=3,
        minimum_channel_threshold=50,
    )
    profiler2.run_one_step()

    profiler2.plot_profiles()
    profiler2.plot_profiles_in_map_view(endpoints_only=True)


def test_different_kwargs(profile_example_grid):
    mg = profile_example_grid
    # with the same grid, test some other profiler options.
    profiler2 = ChannelProfiler(
        mg,
        number_of_watersheds=None,
        main_channel_only=True,
        minimum_outlet_threshold=3,
        minimum_channel_threshold=50,
    )
    profiler2.run_one_step()

    profiler2.plot_profiles()
    profiler2.plot_profiles_in_map_view()

    correct_structure = np.array(
        [
            0,
            61,
            62,
            63,
            64,
            65,
            66,
            67,
            68,
            69,
            70,
            71,
            72,
            73,
            74,
            75,
            76,
            77,
            78,
            79,
            80,
            81,
            82,
            83,
            84,
            85,
            86,
            87,
            88,
            89,
            90,
            91,
            92,
            93,
            94,
            95,
            96,
            97,
            98,
            99,
            100,
            101,
            102,
            103,
            104,
            105,
            106,
            107,
            108,
            109,
        ]
    )
    np.testing.assert_array_equal(profiler2.nodes[0], correct_structure)


def test_re_calculating_nodes_and_distance():
    mg = RasterModelGrid((20, 20), xy_spacing=100)
    z = mg.add_zeros("topographic__elevation", at="node")
    z += np.random.rand(z.size)
    mg.set_closed_boundaries_at_grid_edges(
        bottom_is_closed=False,
        left_is_closed=True,
        right_is_closed=True,
        top_is_closed=True,
    )

    fa = FlowAccumulator(mg, flow_director="D8")
    sp = FastscapeEroder(mg, K_sp=0.0001, m_sp=0.5, n_sp=1)

    dt = 1000
    uplift_per_step = 0.001 * dt

    for i in range(10):
        z[mg.core_nodes] += uplift_per_step
        fa.run_one_step()
        sp.run_one_step(dt=dt)

    profiler = ChannelProfiler(mg)
    profiler.run_one_step()
    assert len(profiler.distance_along_profile) == 1  # result: 1
    profiler.run_one_step()
    # here nathan originally found result: 2, a bug!
    assert len(profiler.distance_along_profile) == 1

    # make the most complicated profile structure
    profiler = ChannelProfiler(mg, main_channel_only=False, number_of_watersheds=2)
    profiler.run_one_step()
    p1 = list(profiler.nodes)
    d1 = list(profiler.distance_along_profile)

    profiler.run_one_step()
    p2 = list(profiler.nodes)
    d2 = list(profiler.distance_along_profile)

    # assert that these are copies, not pointers to same thing
    assert p1 is not p2
    assert d1 is not d2

    # test that structures are the same.
    for idx_watershed in range(len(p1)):
        p1_w = p1[idx_watershed]
        p2_w = p2[idx_watershed]

        d1_w = d1[idx_watershed]
        d2_w = d2[idx_watershed]

        for idx_segment in range(len(p1_w)):
            np.testing.assert_array_equal(p1_w[idx_segment], p2_w[idx_segment])
            np.testing.assert_array_equal(d1_w[idx_segment], d2_w[idx_segment])


@pytest.mark.parametrize("main", [True, False])
@pytest.mark.parametrize("nshed", [1, None, 3])
def test_getting_all_the_way_to_the_divide(main, nshed):
    np.random.seed(42)
    mg = RasterModelGrid((10, 12))
    z = mg.add_zeros("topographic__elevation", at="node")
    z += np.random.rand(z.size)

    fa = FlowAccumulator(mg, flow_director="D8")
    sp = FastscapeEroder(mg, K_sp=0.0001, m_sp=0.5, n_sp=1)

    dt = 1000
    uplift_per_step = 0.001 * dt

    for i in range(100):
        z[mg.core_nodes] += uplift_per_step
        fa.run_one_step()
        sp.run_one_step(dt=dt)

    profiler = ChannelProfiler(
        mg,
        number_of_watersheds=nshed,
        minimum_outlet_threshold=0,
        main_channel_only=main,
        minimum_channel_threshold=0,
    )
    profiler.run_one_step()

    # assert that with minimum_channel_threshold set to zero, we get all the way to the top of the divide.
    for outlet_id in profiler._data_struct:
        seg_tuples = profiler._data_struct[outlet_id].keys()

        wshd_ids = [profiler._data_struct[outlet_id][seg]["ids"] for seg in seg_tuples]

        nodes = np.concatenate(wshd_ids).ravel()
        da = mg.at_node["drainage_area"][nodes]

        # if "profile" is just bits of the edge, then da is 0.
        assert (mg.area_of_cell.min() in da) or (0.0 in da)



import contextlib
import os


@contextlib.contextmanager
def suppress_output(out=True, err=True):
    """Suppress output from both stdout and stderr.

    Parameters
    ----------
    out : bool, optional
        Suppress stdout.
    err : bool, optional
        Suppress stderr.
    """
    null_fds, save_fds = {}, {}
    if out:
        null_fds["out"] = os.open(os.devnull, os.O_RDWR)
        save_fds["out"] = os.dup(1)
        os.dup2(null_fds["out"], 1)
    if err:
        null_fds["err"] = os.open(os.devnull, os.O_RDWR)
        save_fds["err"] = os.dup(2)
        os.dup2(null_fds["err"], 2)

    yield

    # Re-assign the real stdout/stderr back to (1) and (2)
    out and os.dup2(save_fds["out"], 1)
    err and os.dup2(save_fds["err"], 2)

    # Close the null files
    for fd in list(null_fds.values()) + list(save_fds.values()):
        os.close(fd)



import numpy as np
from scipy.integrate import quad

from landlab.components.erosion_deposition.generalized_erosion_deposition import (
    DEFAULT_MINIMUM_TIME_STEP,
    _GeneralizedErosionDeposition,
)
from landlab.utils.return_array import return_array_at_node

from .ext.calc_qs import calculate_qs_in

ROOT2 = np.sqrt(2.0)  # syntactic sugar for precalculated square root of 2
TIME_STEP_FACTOR = 0.5  # factor used in simple subdivision solver


class Space(_GeneralizedErosionDeposition):
    """Stream Power with Alluvium Conservation and Entrainment (SPACE)

    See the publication:
    Shobe, C. M., Tucker, G. E., and Barnhart, K. R.: The SPACE 1.0 model: a
    Landlab component for 2-D calculation of sediment transport, bedrock
    erosion, and landscape evolution, Geosci. Model Dev., 10, 4577-4604,
    `https://doi.org/10.5194/gmd-10-4577-2017 <https://www.geosci-model-dev.net/10/4577/2017/>`_, 2017.

    Unlike other some other fluvial erosion componets in Landlab, in this
    component (and :py:class:`~landlab.components.ErosionDeposition`) no
    erosion occurs in depressions or in areas with adverse slopes. There is no
    ability to pass a keyword argument ``erode_flooded_nodes``.

    If a depressions are handled (as indicated by the presence of the field
    "flood_status_code" at nodes), then deposition occurs throughout the
    depression and sediment is passed out of the depression. Where pits are
    encountered, then all sediment is deposited at that node only.

    Note: If timesteps are large enough that Es*dt (sediment erosion)
    exceeds sediment thickness H, the 'adaptive' solver is necessary to
    subdivide timesteps. Compare Es and H arrays to determine whether
    timesteps are appropriate or too large for the 'basic' solver.

    Examples
    ---------
    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> from landlab.components import (FlowAccumulator,
    ...                                 DepressionFinderAndRouter,
    ...                                 Space,
    ...                                 FastscapeEroder)
    >>> np.random.seed(seed = 5000)

    Define grid and initial topography:

    *  5x5 grid with baselevel in the lower left corner
    *  All other boundary nodes closed
    *  Initial topography is plane tilted up to the upper right with
       noise

    >>> mg = RasterModelGrid((5, 5), xy_spacing=10.0)
    >>> _ = mg.add_zeros('topographic__elevation', at='node')
    >>> mg.at_node['topographic__elevation'] += (mg.node_y / 10. +
    ...     mg.node_x / 10. + np.random.rand(len(mg.node_y)) / 10.)
    >>> mg.set_closed_boundaries_at_grid_edges(bottom_is_closed=True,
    ...                                        left_is_closed=True,
    ...                                        right_is_closed=True,
    ...                                        top_is_closed=True)
    >>> mg.set_watershed_boundary_condition_outlet_id(
    ...     0, mg.at_node['topographic__elevation'], -9999.)
    >>> fsc_dt = 100.
    >>> space_dt = 100.

    Instantiate Fastscape eroder, flow router, and depression finder

    >>> fr = FlowAccumulator(mg, flow_director='D8')
    >>> df = DepressionFinderAndRouter(mg)
    >>> fsc = FastscapeEroder(
    ...     mg,
    ...     K_sp=.001,
    ...     m_sp=.5,
    ...     n_sp=1)

    Burn in an initial drainage network using the Fastscape eroder:

    >>> for x in range(100):
    ...     fr.run_one_step()
    ...     df.map_depressions()
    ...     fsc.run_one_step(dt=fsc_dt)
    ...     mg.at_node['topographic__elevation'][0] -= 0.001 # Uplift

    Add some soil to the drainage network:

    >>> _ = mg.add_zeros('soil__depth', at='node', dtype=float)
    >>> mg.at_node['soil__depth'] += 0.5
    >>> mg.at_node['topographic__elevation'] += mg.at_node['soil__depth']

    Instantiate the Space component:

    >>> ha = Space(
    ...     mg,
    ...     K_sed=0.00001,
    ...     K_br=0.00000000001,
    ...     F_f=0.5,
    ...     phi=0.1,
    ...     H_star=1.,
    ...     v_s=0.001,
    ...     m_sp=0.5,
    ...     n_sp = 1.0,
    ...     sp_crit_sed=0,
    ...     sp_crit_br=0)

    Now run the Space component for 2000 short timesteps:

    >>> for x in range(2000): #Space component loop
    ...     fr.run_one_step()
    ...     df.map_depressions()
    ...     ha.run_one_step(dt=space_dt)
    ...     mg.at_node['bedrock__elevation'][0] -= 2e-6 * space_dt

    Now we test to see if soil depth and topography are right:

    >>> np.around(mg.at_node['soil__depth'], decimals=3) # doctest: +NORMALIZE_WHITESPACE
    array([ 0.5  ,  0.5  ,  0.5  ,  0.5  ,  0.5  ,  0.5  ,  0.495,  0.492,
            0.491,  0.5  ,  0.5  ,  0.492,  0.492,  0.49 ,  0.5  ,  0.5  ,
            0.491,  0.49 ,  0.484,  0.5  ,  0.5  ,  0.5  ,  0.5  ,  0.5  ,
            0.5  ])

    >>> np.around(mg.at_node['topographic__elevation'], decimals=3) # doctest: +NORMALIZE_WHITESPACE
    array([ 0.423,  1.536,  2.573,  3.511,  4.561,  1.582,  0.424,  0.428,
            0.438,  5.51 ,  2.54 ,  0.428,  0.428,  0.438,  6.526,  3.559,
            0.438,  0.438,  0.45 ,  7.553,  4.559,  5.541,  6.57 ,  7.504,
            8.51 ])

    References
    ----------
    **Required Software Citation(s) Specific to this Component**

    Shobe, C., Tucker, G., Barnhart, K. (2017). The SPACE 1.0 model: a Landlab
    component for 2-D calculation of sediment transport, bedrock erosion, and
    landscape evolution. Geoscientific Model Development  10(12), 4577 - 4604.
    https://dx.doi.org/10.5194/gmd-10-4577-2017

    **Additional References**

    None Listed

    """

    _name = "Space"

    _unit_agnostic = True

    _info = {
        "flow__link_to_receiver_node": {
            "dtype": int,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "ID of link downstream of each node, which carries the discharge",
        },
        "flow__receiver_node": {
            "dtype": int,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "Node array of receivers (node that receives flow from current node)",
        },
        "flow__upstream_node_order": {
            "dtype": int,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "Node array containing downstream-to-upstream ordered list of node IDs",
        },
        "sediment__flux": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m3/s",
            "mapping": "node",
            "doc": "Sediment flux (volume per unit time of sediment entering each node)",
        },
        "soil__depth": {
            "dtype": float,
            "intent": "inout",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "Depth of soil or weathered bedrock",
        },
        "surface_water__discharge": {
            "dtype": float,
            "intent": "in",
            "optional": False,
            "units": "m**3/s",
            "mapping": "node",
            "doc": "Volumetric discharge of surface water",
        },
        "topographic__elevation": {
            "dtype": float,
            "intent": "inout",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "Land surface topographic elevation",
        },
        "topographic__steepest_slope": {
            "dtype": float,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "The steepest *downhill* slope",
        },
    }

    _cite_as = """@Article{gmd-10-4577-2017,
                  AUTHOR = {Shobe, C. M. and Tucker, G. E. and Barnhart, K. R.},
                  TITLE = {The SPACE~1.0 model: a~Landlab component for 2-D calculation of sediment transport, bedrock erosion, and landscape evolution},
                  JOURNAL = {Geoscientific Model Development},
                  VOLUME = {10},
                  YEAR = {2017},
                  NUMBER = {12},
                  PAGES = {4577--4604},
                  URL = {https://www.geosci-model-dev.net/10/4577/2017/},
                  DOI = {10.5194/gmd-10-4577-2017}
                  }"""

    def __init__(
        self,
        grid,
        K_sed=0.002,
        K_br=0.002,
        F_f=0.0,
        phi=0.3,
        H_star=0.1,
        v_s=1.0,
        m_sp=0.5,
        n_sp=1.0,
        sp_crit_sed=0.0,
        sp_crit_br=0.0,
        discharge_field="surface_water__discharge",
        solver="basic",
        dt_min=DEFAULT_MINIMUM_TIME_STEP,
    ):
        """Initialize the Space model.

        Parameters
        ----------
        grid : ModelGrid
            Landlab ModelGrid object
        K_sed : float, field name, or array
            Erodibility for sediment (units vary).
        K_br : float, field name, or array
            Erodibility for bedrock (units vary).
        F_f : float
            Fraction of permanently suspendable fines in bedrock [-].
        phi : float
            Sediment porosity [-].
        H_star : float
            Sediment thickness required for full entrainment [L].
        v_s : float
            Effective settling velocity for chosen grain size metric [L/T].
        m_sp : float
            Drainage area exponent (units vary)
        n_sp : float
            Slope exponent (units vary)
        sp_crit_sed : float, field name, or array
            Critical stream power to erode sediment [E/(TL^2)]
        sp_crit_br : float, field name, or array
            Critical stream power to erode rock [E/(TL^2)]
        discharge_field : float, field name, or array
            Discharge [L^2/T]. The default is to use the grid field
            'surface_water__discharge', which is simply drainage area
            multiplied by the default rainfall rate (1 m/yr). To use custom
            spatially/temporally varying rainfall, use 'water__unit_flux_in'
            to specify water input to the FlowAccumulator.
        solver : string
            Solver to use. Options at present include:
                (1) 'basic' (default): explicit forward-time extrapolation.
                    Simple but will become unstable if time step is too large.
                (2) 'adaptive': subdivides global time step as needed to
                    prevent slopes from reversing and alluvium from going
                    negative.

        """
        if grid.at_node["flow__receiver_node"].size != grid.size("node"):
            msg = (
                "A route-to-multiple flow director has been "
                "run on this grid. The landlab development team has not "
                "verified that SPACE is compatible with "
                "route-to-multiple methods. Please open a GitHub Issue "
                "to start this process."
            )
            raise NotImplementedError(msg)

        super().__init__(
            grid,
            m_sp=m_sp,
            n_sp=n_sp,
            F_f=F_f,
            v_s=v_s,
            dt_min=dt_min,
            discharge_field=discharge_field,
        )

        if phi >= 1.0:
            raise ValueError("Porosity must be < 1.0")
        if phi < 0.0:
            raise ValueError("Porosity must be > 0.0")

        self._phi = float(phi)
        self._porosity_factor = 1.0 / (1.0 - self._phi)

        # space specific inits
        self._H_star = H_star
        self._soil__depth = grid.at_node["soil__depth"]

        if "bedrock__elevation" in grid.at_node:
            self._bedrock__elevation = grid.at_node["bedrock__elevation"]
        else:
            self._bedrock__elevation = grid.add_zeros(
                "bedrock__elevation", at="node", dtype=float
            )

            self._bedrock__elevation[:] = (
                self._topographic__elevation - self._soil__depth
            )

        self._sed_erosion_term = np.zeros(grid.number_of_nodes)
        self._br_erosion_term = np.zeros(grid.number_of_nodes)
        self._Es = np.zeros(grid.number_of_nodes)
        self._Er = np.zeros(grid.number_of_nodes)

        # K's and critical values can be floats, grid fields, or arrays
        # use setters defined below
        self.K_sed = K_sed
        self.K_br = K_br

        self._sp_crit_sed = return_array_at_node(grid, sp_crit_sed)
        self._sp_crit_br = return_array_at_node(grid, sp_crit_br)

        # Handle option for solver
        if solver == "basic":
            self.run_one_step = self.run_one_step_basic
        elif solver == "adaptive":
            self.run_one_step = self.run_with_adaptive_time_step_solver
            self._time_to_flat = np.zeros(grid.number_of_nodes)
            self._time_to_zero_alluv = np.zeros(grid.number_of_nodes)
            self._dzdt = np.zeros(grid.number_of_nodes)
        else:
            raise ValueError(
                "Parameter 'solver' must be one of: " + "'basic', 'adaptive'"
            )

    @property
    def K_br(self):
        """Erodibility of bedrock(units depend on m_sp)."""
        return self._K_br

    @K_br.setter
    def K_br(self, new_val):
        self._K_br = return_array_at_node(self._grid, new_val)

    @property
    def K_sed(self):
        """Erodibility of sediment(units depend on m_sp)."""
        return self._K_sed

    @K_sed.setter
    def K_sed(self, new_val):
        self._K_sed = return_array_at_node(self._grid, new_val)

    def _calc_erosion_rates(self):
        """Calculate erosion rates."""
        # if sp_crits are zero, then this colapses to correct all the time.
        if self._n_sp == 1.0:
            S_to_the_n = self._slope
        else:
            S_to_the_n = np.power(self._slope, self._n_sp)
        omega_sed = self._K_sed * self._Q_to_the_m * S_to_the_n
        omega_br = self._K_br * self._Q_to_the_m * S_to_the_n

        omega_sed_over_sp_crit = np.divide(
            omega_sed,
            self._sp_crit_sed,
            out=np.zeros_like(omega_sed),
            where=self._sp_crit_sed != 0,
        )

        omega_br_over_sp_crit = np.divide(
            omega_br,
            self._sp_crit_br,
            out=np.zeros_like(omega_br),
            where=self._sp_crit_br != 0,
        )

        self._sed_erosion_term = omega_sed - self._sp_crit_sed * (
            1.0 - np.exp(-omega_sed_over_sp_crit)
        ) / (
            1 - self._phi
        )  # convert from a volume to a mass flux.

        self._br_erosion_term = omega_br - self._sp_crit_br * (
            1.0 - np.exp(-omega_br_over_sp_crit)
        )

        H_over_Hstar = self._soil__depth / self._H_star
        self._Es = self._sed_erosion_term * (1.0 - np.exp(-H_over_Hstar))

        self._Er = self._br_erosion_term * np.exp(-H_over_Hstar)

    @property
    def Es(self):
        """Sediment erosion term."""
        return self._Es

    @property
    def Er(self):
        """Bedrock erosion term."""
        return self._Er

    @property
    def H(self):
        """Sediment thickness."""
        return self._H

    def _calc_qs_in_and_depo_rate(self):

        # Choose a method for calculating erosion:
        self._calc_hydrology()
        self._calc_erosion_rates()

        is_flooded_core_node = self._get_flooded_core_nodes()

        self._Es[is_flooded_core_node] = 0.0
        self._Er[is_flooded_core_node] = 0.0

        self._sed_erosion_term[is_flooded_core_node] = 0.0
        self._br_erosion_term[is_flooded_core_node] = 0.0

        self._qs_in[:] = 0
        self._depo_rate[:] = 0.0

        # iterate top to bottom through the stack, calculate qs
        # cythonized version of calculating qs_in
        calculate_qs_in(
            np.flipud(self._stack),
            self._flow_receivers,
            self._cell_area_at_node,
            self._q,
            self._qs,
            self._qs_in,
            self._Es,
            self._Er,
            self._v_s,
            self._F_f,
        )

        self._depo_rate[self._q > 0] = self._qs[self._q > 0] * (
            self._v_s / self._q[self._q > 0]
        )

        if not self._depressions_are_handled():  # all sed dropped here
            self._depo_rate[is_flooded_core_node] = (
                self._qs_in[is_flooded_core_node]
                / self._cell_area_at_node[is_flooded_core_node]
            )
        return is_flooded_core_node

    def run_one_step_basic(self, dt=1.0):
        """Calculate change in rock and alluvium thickness for a time period
        'dt'.

        Parameters
        ----------
        dt : float
            Model timestep [T]
        """
        self._calc_qs_in_and_depo_rate()
        cores = self._grid.core_nodes

        H0 = self._soil__depth.copy()

        # now, the analytical solution to soil thickness in time:
        # need to distinguish D=kqS from all other cases to save from blowup!

        # distinguish cases:

        # there is a numerical blow up as well, which is not discussed in the
        # paper. when H>>H* the analytical solution (Eq 34) has a term that is
        # exp (H/H*) This can becom infinite. We will consider the case of
        # H/H*>50 as distinct.
        H_over_H_star = self._soil__depth / self._H_star
        too_thick = H_over_H_star > 100

        H_over_H_star[too_thick] = 100

        no_entrainment = self._sed_erosion_term <= 0.0  # this will include pits.
        blowup = (
            (self._depo_rate == self._sed_erosion_term)
            & (self._sed_erosion_term > 0.0)
            & (~no_entrainment)
        )
        full_equation = (~blowup) & (~no_entrainment)

        # First do blowup.
        # this is Space paper Eq 34. This example has the same issues with
        # very thick sed. If sed is very thick AND ero=depo, there is no change.
        self._soil__depth[blowup * (~too_thick)] = self._H_star * np.log(
            ((self._sed_erosion_term[blowup]) / self._H_star) * dt
            + np.exp(H_over_H_star[blowup])
        )

        # Second, do no entrainment of sediment. This is equation 35.
        self._soil__depth[no_entrainment] += (
            self._depo_rate[no_entrainment] / (1 - self._phi)
        ) * dt

        # Treat the case of very thick sediments and ero != depo.
        self._soil__depth[full_equation * too_thick] += (
            (
                self._depo_rate[full_equation * (full_equation * too_thick)]
                / (1 - self._phi)
            )
            - (self._sed_erosion_term[full_equation * too_thick] / (1 - self._phi))
        ) * dt

        # Finally do the full equation (Eq 32) when not too thick.
        self._soil__depth[full_equation * (~too_thick)] = self._H_star * np.log(
            (
                1
                / (
                    (self._depo_rate[full_equation * (~too_thick)] / (1 - self._phi))
                    / (
                        self._sed_erosion_term[full_equation * (~too_thick)]
                        / (1 - self._phi)
                    )
                    - 1
                )
            )
            * (
                np.exp(
                    (
                        self._depo_rate[full_equation * (~too_thick)] / (1 - self._phi)
                        - (
                            self._sed_erosion_term[full_equation * (~too_thick)]
                            / (1 - self._phi)
                        )
                    )
                    * (dt / self._H_star)
                )
                * (
                    (
                        (
                            self._depo_rate[full_equation * (~too_thick)]
                            / (1 - self._phi)
                            / (
                                self._sed_erosion_term[full_equation * (~too_thick)]
                                / (1 - self._phi)
                            )
                        )
                        - 1
                    )
                    * np.exp(H_over_H_star[full_equation * (~too_thick)])
                    + 1
                )
                - 1
            )
        )

        # Equation 12 gives dRdt, and Equation 36 gives R. However, we don't
        # include dH/dt within timestep in integrating for R.
        # This matters when we are starting with very little soil and increasing.

        # to do this right I think we need a numerical integral for
        # R(t). Unfortunately we have three cases for H(t). These are now
        # implemented in the function _dRdt. This loop and the solver can be
        # cythonized. But not today.

        dR = self._grid.zeros(at="node")
        for idx in self.grid.core_nodes:
            args = (
                self._br_erosion_term[idx],
                1.0 / self._H_star,
                self._depo_rate[idx] / (1.0 - self._phi),
                self._sed_erosion_term[idx] / (1.0 - self._phi),
                H0[idx],
            )

            dR[idx] = quad(_dRdt, 0, dt, args)[0]

        self._bedrock__elevation += dR

        # finally, determine topography by summing bedrock and soil
        self._topographic__elevation[cores] = (
            self._bedrock__elevation[cores] + self._soil__depth[cores]
        )

    def run_with_adaptive_time_step_solver(self, dt=1.0):
        """Run step with CHILD-like solver that adjusts time steps to prevent
        slope flattening.

        Parameters
        ----------
        dt : float
            Model timestep [T]

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> from landlab.components import FlowAccumulator
        >>> import numpy as np

        >>> rg = RasterModelGrid((3, 4))
        >>> z = rg.add_zeros('topographic__elevation', at='node')
        >>> z[:] = 0.1 * rg.x_of_node
        >>> H = rg.add_zeros('soil__depth', at='node')
        >>> H += 0.1
        >>> br = rg.add_zeros('bedrock__elevation', at='node')
        >>> br[:] = z - H

        >>> fa = FlowAccumulator(rg, flow_director='FlowDirectorSteepest')
        >>> fa.run_one_step()
        >>> sp = Space(rg, K_sed=1.0, K_br=0.1,
        ...            F_f=0.5, phi=0.0, H_star=1., v_s=1.0,
        ...            m_sp=0.5, n_sp = 1.0, sp_crit_sed=0,
        ...            sp_crit_br=0, solver='adaptive')
        >>> sp.run_one_step(dt=10.0)

        >>> np.round(sp.Es[5:7], 4)
        array([ 0.0029,  0.0074])
        >>> np.round(sp.Er[5:7], 4)
        array([ 0.0032,  0.0085])
        >>> np.round(H[5:7], 3)
        array([ 0.088,  0.078])
        """

        # Initialize remaining_time, which records how much of the global time
        # step we have yet to use up.
        remaining_time = dt

        z = self._grid.at_node["topographic__elevation"]
        br = self._grid.at_node["bedrock__elevation"]
        H = self._grid.at_node["soil__depth"]
        r = self._flow_receivers
        cores = self._grid.core_nodes

        first_iteration = True

        is_flooded_core_node = self._get_flooded_core_nodes()

        # Outer WHILE loop: keep going until time is used up
        while remaining_time > 0.0:

            # Update all the flow-link slopes.
            #
            # For the first iteration, we assume this has already been done
            # outside the component (e.g., by flow router), but we need to do
            # it ourselves on subsequent iterations.
            if not first_iteration:
                # update the link slopes
                self._update_flow_link_slopes()
                # update where nodes are flooded. This shouuldn't happen because
                # of the dynamic timestepper, but just in case, we update here.
                is_flooded_core_node[self._slope < 0] = True
            else:
                first_iteration = False

            is_flooded_core_node = (
                self._calc_qs_in_and_depo_rate()
            )  # THIS IS THE SPEED BOTTLENECK

            # Now look at upstream-downstream node pairs, and recording the
            # time it would take for each pair to flatten. Take the minimum.
            self._dzdt[cores] = self._depo_rate[cores] * self._porosity_factor - (
                self._Es[cores] + self._Er[cores]
            )
            rocdif = self._dzdt - self._dzdt[r]
            zdif = z - z[r]
            self._time_to_flat[:] = remaining_time

            converging = np.where(rocdif < 0.0)[0]
            self._time_to_flat[converging] = -(
                TIME_STEP_FACTOR * zdif[converging] / rocdif[converging]
            )
            self._time_to_flat[np.where(zdif <= 0.0)[0]] = remaining_time

            # From this, find the maximum stable time step with regard to slope
            # evolution.
            dt_max1 = np.amin(self._time_to_flat)

            # Next we consider time to exhaust regolith
            self._time_to_zero_alluv[:] = remaining_time

            # poof deposition by phi
            dHdt = self._porosity_factor * (self._depo_rate - self._Es)
            decreasing_H = np.where(dHdt < 0.0)[0]
            self._time_to_zero_alluv[decreasing_H] = -(
                TIME_STEP_FACTOR * H[decreasing_H] / dHdt[decreasing_H]
            )

            # Now find the smallest time that would lead to near-empty alluv
            dt_max2 = np.amin(self._time_to_zero_alluv)

            # Take the smaller of the limits
            dt_max = max(self._dt_min, min(dt_max1, dt_max2))

            # Now a vector operation: apply dzdt and dhdt to all nodes
            br[cores] -= self._Er[cores] * dt_max
            H[cores] += dHdt[cores] * dt_max
            z[cores] = br[cores] + H[cores]

            # Update remaining time and continue
            remaining_time -= dt_max


def _dRdt(t, a, b, c, d, H0):
    """
    dRdt = a * exp(-b * H (t))

    a =  Kr q S*n
    b = 1/H*

    c = VQs/(Q * (1-phi))
    d = Ks q Sn

    if d <= 0:
        H (t) = (1/b) ln [ d*b*t + exp(H0/H*) ]

    """
    # truncate H/H*
    if b * H0 > 100:
        too_thick = True
    else:
        too_thick = False
    bH0 = min(b * H0, 100)

    # set cases for H depending on values of constants.

    # deposition only case.
    if d <= 0:
        H = H0 + (c * t)
    else:
        # sediment deposition = sediment entrainment.
        if c == d:
            if too_thick:
                H = H0
            else:
                H = (1 / b) * np.log((d * b * t) + np.exp(bH0))
        # full equation 32
        else:
            if too_thick:
                H = H0 + (c - d) * t
            else:
                term1 = 1.0 / ((c / d) - 1)
                term2 = np.exp((c - d) * t * b)
                term3 = (c / d - 1) * np.exp(bH0) + 1
                interior = term1 * (term2 * (term3) - 1)
                H = (1 / b) * np.log(interior)

    # calculate dRdt
    dRdt = -a * np.exp(-b * H)
    return dRdt



import numpy as np
import pytest

from landlab import RasterModelGrid
from landlab.components import FlowAccumulator, HeightAboveDrainageCalculator


def test_route_to_multiple_error_raised():
    mg = RasterModelGrid((10, 10))
    z = mg.add_zeros("topographic__elevation", at="node")
    z += mg.x_of_node + mg.y_of_node
    fa = FlowAccumulator(mg, flow_director="MFD")
    fa.run_one_step()

    channel__mask = mg.zeros(at="node")

    with pytest.raises(NotImplementedError):
        HeightAboveDrainageCalculator(mg, channel_mask=channel__mask)


def test_warn_drainage_pits():
    mg = RasterModelGrid((4, 4))
    z = mg.add_zeros("topographic__elevation", at="node")
    elev = np.array([[2, 1, 1, 2], [3, 2, 2, 3], [4, 1, 3, 4], [5, 3, 4, 4]])
    z[:] = elev.reshape(len(z))

    fa = FlowAccumulator(mg, flow_director="D8")
    fa.run_one_step()

    channel__mask = mg.zeros(at="node")
    channel__mask[[2, 6]] = 1
    hd = HeightAboveDrainageCalculator(mg, channel_mask=channel__mask)

    with pytest.warns(UserWarning):
        hd.run_one_step()
